# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "26/04/2021"

# import os

import logging
import numpy

from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.StackView import StackViewMainWindow
from silx.gui.plot.items.roi import RectangleROI
from silx.gui.plot.tools.roi import RegionOfInterestManager, RegionOfInterestTableWidget

import darfix
from darfix.core.dataset import Operation
from .operationThread import OperationThread


_logger = logging.getLogger(__file__)


class ROISelectionWidget(qt.QWidget):
    """
    Widget that allows the user to pick a ROI in any image of the dataset.
    """

    sigComputed = qt.Signal(list, list)

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self.roi = None
        self._update_dataset = None
        self.indices = None
        self.bg_indices = None
        self.bg_dataset = None
        self._original_dataset = True

        self.setLayout(qt.QVBoxLayout())
        self._sv = StackViewMainWindow()
        _buttons = qt.QDialogButtonBox(parent=self)
        self._okB = _buttons.addButton(_buttons.Ok)
        self._applyB = _buttons.addButton(_buttons.Apply)
        self._abortB = _buttons.addButton(_buttons.Abort)
        self._resetB = _buttons.addButton(_buttons.Reset)
        self._abortB.hide()

        self._applyB.clicked.connect(self.applyRoi)
        self._okB.clicked.connect(self.apply)
        self._resetB.clicked.connect(self.resetStack)
        self._abortB.clicked.connect(self.abort)

        self._sv.setColormap(
            Colormap(
                name=darfix.config.DEFAULT_COLORMAP_NAME,
                normalization=darfix.config.DEFAULT_COLORMAP_NORM,
            )
        )
        self.layout().addWidget(self._sv)
        self.layout().addWidget(_buttons)

        plot = self._sv.getPlot()
        self._roiManager = RegionOfInterestManager(plot)
        self._roiTable = RegionOfInterestTableWidget()
        self._roiTable.setRegionOfInterestManager(self._roiManager)

        self._roi = RectangleROI()
        self._roi.setLabel("ROI")
        self._roi.setGeometry(origin=(0, 0), size=(10, 10))
        self._roi.setEditable(True)
        self._roiManager.addRoi(self._roi)
        self._roiTable.setColumnHidden(4, True)

        # Add the region of interest table and the buttons to a dock widget
        widget = qt.QWidget()
        layout = qt.QVBoxLayout()
        widget.setLayout(layout)
        layout.addWidget(self._roiTable)

        def roiDockVisibilityChanged(visible):
            """Handle change of visibility of the roi dock widget.

            If dock becomes hidden, ROI interaction is stopped.
            """
            if not visible:
                self._roiManager.stop()

        dock = qt.QDockWidget("Image ROI")
        dock.setWidget(widget)
        dock.visibilityChanged.connect(roiDockVisibilityChanged)
        plot.addTabbedDockWidget(dock)

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter. Saves the dataset and updates the stack with the dataset
        data

        :param Dataset dataset: dataset
        """
        self._parent = parent
        self._dataset = dataset
        self._update_dataset = dataset
        self.indices = indices
        if self._dataset.title != "":
            self._sv.setTitleCallback(lambda idx: self._dataset.title)
        self.setStack()
        self.bg_indices = bg_indices
        self.bg_dataset = bg_dataset

    def _updateDataset(self, widget, dataset):
        self._dataset = dataset
        self._update_dataset = dataset
        self._parent._updateDataset(widget, dataset)

    def setStack(self, dataset=None):
        """
        Sets new data to the stack.
        Mantains the current frame showed in the view.

        :param Dataset dataset: if not None, data set to the stack will be from the given dataset.
        """
        if dataset is None:
            dataset = self._dataset
        first_frame_shape = dataset.get_data()[0].shape
        self.setRoi(
            center=(first_frame_shape[1] / 2, first_frame_shape[0] / 2),
            size=(first_frame_shape[1] / 5, first_frame_shape[0] / 5),
        )
        nframe = self._sv.getFrameNumber()
        self._sv.setStack(dataset.get_data())
        self._sv.setFrameNumber(nframe)

    def setRoi(self, roi=None, origin=None, size=None, center=None):
        """
        Sets a region of interest of the stack of images.

        :param RectangleROI roi: A region of interest.
        :param Tuple origin: If a roi is not provided, used as an origin for the roi
        :param Tuple size: If a roi is not provided, used as a size for the roi.
        :param Tuple center: If a roi is not provided, used as a center for the roi.
        """
        if roi is not None and (
            size is not None or center is not None or origin is not None
        ):
            _logger.warning(
                "Only using provided roi, the rest of parameters are omitted"
            )

        if roi is not None:
            self._roi = roi
        else:
            self._roi.setGeometry(origin=origin, size=size, center=center)

    def getRoi(self):
        """
        Returns the roi selected in the stackview.

        :rtype: silx.gui.plot.items.roi.RectangleROI
        """
        return self._roi

    def applyRoi(self):
        """
        Function to apply the region of interest at the data of the dataset
        and show the new data in the stack. Dataset data is not yet replaced.
        A new roi is created in the middle of the new stack.
        """
        if not self._update_dataset.in_memory:
            self._abortB.show()
        self._applyB.setEnabled(False)
        self._okB.setEnabled(False)
        self.roi = RectangleROI()
        self.roi.setGeometry(
            origin=self.getRoi().getOrigin(), size=self.getRoi().getSize()
        )
        self.thread = OperationThread(self, self._update_dataset.apply_roi)
        roi_dir = self._update_dataset.dir if not self._original_dataset else None
        self.thread.setArgs(
            size=numpy.flip(self.roi.getSize()),
            center=numpy.flip(self.roi.getCenter()),
            roi_dir=roi_dir,
        )
        self.thread.finished.connect(self._updateData)
        self.thread.start()

    def abort(self):
        self._abortB.setEnabled(False)
        self._update_dataset.stop_operation(Operation.ROI)

    def _updateData(self):
        """
        Updates the stack with the data computed in the thread
        """
        self.thread.finished.disconnect(self._updateData)
        self._abortB.hide()
        self._abortB.setEnabled(True)
        self._applyB.setEnabled(True)
        self._okB.setEnabled(True)
        if self.thread.data:
            self._update_dataset = self.thread.data
            self.thread.func = None
            self.thread.data = None
            self._original_dataset = False
            assert self._update_dataset is not None
            self.setStack(self._update_dataset)
            self.resetROI()
        else:
            print("\nCorrection aborted")

    def apply(self):
        """
        Function that replaces the dataset data with the data shown in the stack of images.
        If the stack has a roi applied, it applies the same roi to the dark frames of the dataset.
        Signal emitted with the roi parameters.
        """
        if self.roi:
            self.sigComputed.emit(
                self.roi.getOrigin().tolist(), self.roi.getSize().tolist()
            )
        else:
            self.sigComputed.emit([], [])

    def getDataset(self):
        if self.roi:
            bg_dataset = (
                self.bg_dataset.apply_roi(
                    size=numpy.flip(self.roi.getSize()),
                    center=numpy.flip(self.roi.getCenter()),
                )
                if self.bg_dataset is not None
                else None
            )
        else:
            bg_dataset = self.bg_dataset
        return self._update_dataset, self.indices, self.bg_indices, bg_dataset

    def getStackViewColormap(self):
        """
        Returns the colormap from the stackView

        :rtype: silx.gui.colors.Colormap
        """
        return self._sv.getColormap()

    def setStackViewColormap(self, colormap):
        """
        Sets the stackView colormap

        :param colormap: Colormap to set
        :type colormap: silx.gui.colors.Colormap
        """
        self._sv.setColormap(colormap)

    def resetROI(self):
        """
        Sets the region of interest in the middle of the stack, with size 1/5 of the image.
        """
        frame_shape = numpy.array(self._update_dataset.get_data(0).shape)
        center = numpy.flip(frame_shape) / 2
        size = numpy.flip(frame_shape) / 5
        self.setRoi(center=center, size=size)

    def resetStack(self):
        """
        Restores stack with the dataset data.
        """
        self.roi = None
        self._update_dataset = self._dataset
        self._original_dataset = True
        self.setStack(self._dataset)

    def clearStack(self):
        """
        Clears stack.
        """
        self._sv.setStack(None)
        self._roi.setGeometry(origin=(0, 0), size=(10, 10))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "12/06/2020"

import json
import numpy

from silx.io import fabioh5
from silx.io.url import DataUrl

from darfix.core.dataset import Data, Dataset
from darfix.core.dimension import AcquisitionDims


def save_to_json(
    filename, dataset, original_dataset=None, hi_indices=None, li_indices=None
):
    my_dict = {}
    my_dict["dir"] = dataset.dir
    my_dict["dataset"] = [url.file_path() for url in dataset.get_data().urls]
    my_dict["shape"] = dataset.data.shape
    my_dict["in_memory"] = dataset.in_memory
    if original_dataset is not None:
        my_dict["original_dataset"] = [
            url.file_path() for url in original_dataset.get_data().urls
        ]
    if hi_indices is not None:
        my_dict["hi_indices"] = hi_indices.tolist()
    if li_indices is not None:
        my_dict["li_indices"] = li_indices.tolist()
    if dataset.dims.ndim > 0:
        my_dict["dims"] = dataset.dims.to_dict()
    with open(filename + ".json", "w") as f:
        json.dump(my_dict, f)


def load_from_json(filename):
    hi_indices = li_indices = dims = None
    with open(filename + ".json", "r") as f:
        distro = json.load(f)
        urls = numpy.array(
            [DataUrl(file_path=url, scheme="fabio") for url in distro["dataset"]]
        )
        if "original_dataset" in distro:
            original_dataset = distro["original_dataset"]
        else:
            original_dataset = distro["dataset"]
        metadata = []
        for filename in original_dataset:
            fabio_reader = fabioh5.EdfFabioReader(file_name=filename)
            metadata.append(fabio_reader)
            fabio_reader.close()
        metadata = numpy.array(metadata)
        hi_indices = distro["hi_indices"] if "hi_indices" in distro else None
        li_indices = distro["li_indices"] if "li_indices" in distro else None
        if "dims" in distro:
            dims = AcquisitionDims()
            dims.from_dict(distro["dims"])
        else:
            dims = None
        data = Data(urls, metadata, distro["in_memory"]).reshape(distro["shape"])
        dataset = Dataset(
            _dir=distro["dir"], data=data, in_memory=distro["in_memory"], dims=dims
        )

    return dataset, hi_indices, li_indices



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "19/03/2020"

import numpy

from .base import Base


class PCA(Base):
    def __init__(
        self, data, num_components=None, center=True, whiten=False, rowvar=True
    ):
        Base.__init__(self, data, num_components=num_components)

        self.rowvar = rowvar
        if self.rowvar:
            self.data = self.data[:, :].T

        self._rows, self._columns = self.data.shape
        self._data_orig = self.data
        self._mean = self._data_orig[:, :].mean(axis=0)

        self._whiten = whiten
        self._center = center

        if self._center:
            # copy the data before centering it
            self.data = self._data_orig - self._mean

    def _update_h(self):

        self.H = numpy.matmul(self.data[:, :], self.W.T)

        if self.rowvar:
            self.H = self.H.T

    def _update_w(self):

        X = self.data[:, :]
        r = self.num_components

        if self._columns > self._rows:  # n x n matrix
            if not self._center:
                M = numpy.matmul((X - self._mean), (X - self._mean).T)
            else:
                M = numpy.matmul(X, X.T)

            # Eigenvector decomposition
            vals, vecs = numpy.linalg.eig(M)
            vals, vecs = vals.real, vecs.real

            # Sort the eigenvectors by "importance" and get the first r
            pairs = sorted(
                [(vals[i], vecs[:, i]) for i in range(len(vals))],
                key=lambda x: x[0],
                reverse=True,
            )
            pairs = [
                p for p in pairs if abs(p[0]) > 1e-10
            ]  # Remove the eigenvectors of 0 eigenvalue
            pairs = pairs[:r]

            # nxr matrix of eigenvectors (each column is an n-dimensional eigenvector)
            E = numpy.array([p[1] for p in pairs]).transpose()

            # pxr matrix of the first r eigenvectors of the covariance of X
            # Note that we normalize!
            E = numpy.matmul((X - self._mean).transpose(), E)
            E /= numpy.linalg.norm(E, axis=0)

            if self._whiten:
                # Eigenvalues of cov(X) to the -1/2
                # Note that we rescale the eigenvalues of M to get the eigenvalues of cov(X)!
                diag = numpy.array(
                    [1 / numpy.sqrt(p[0] / (self._rows - 1)) for p in pairs]
                )

        else:  # p x p matrix
            C = numpy.cov(X, rowvar=False)

            # Eigenvector decomposition
            vals, vecs = numpy.linalg.eig(C)
            vals, vecs = vals.real, vecs.real

            # Sort the eigenvectors by "importance" and get the first r
            pairs = sorted(
                [(vals[i], vecs[:, i]) for i in range(len(vals))],
                key=lambda x: x[0],
                reverse=True,
            )
            pairs = [
                p for p in pairs if abs(p[0]) > 1e-10
            ]  # Remove the eigenvectors of 0 eigenvalue
            pairs = pairs[:]

            # pxr matrix of the first r eigenvectors of the covariance of X
            E = numpy.array([p[1] for p in pairs]).transpose()

            if self._whiten:  # Eigenvalues of cov(X) to the -1/2
                diag = numpy.array([1 / numpy.sqrt(p[0]) for p in pairs])

        if self._whiten:
            self.W = E * diag
        else:
            self.W = E

        self.W = self.W.T

    def fit_transform(self, max_iter=1, error_step=None):

        Base.fit_transform(self, max_iter=1, error_step=error_step)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "10/08/2021"


from Orange.widgets.settings import Setting
from silx.gui import qt
from Orange.widgets.widget import OWWidget, Input, Output
from darfix.gui.magnificationWidget import MagnificationWidget
from darfix.gui.rsmWidget import RSMWidget
from darfix.core.process import Transformation


class TransformationWidgetOW(OWWidget):
    """
    Widget that computes the background substraction from a dataset
    """

    name = "transformation"
    icon = "icons/axes.png"
    want_main_area = False
    ewokstaskclass = Transformation

    # Inputs
    class Inputs:
        dataset = Input("dataset", tuple)

    # Outputs
    class Outputs:
        dataset = Output("dataset", tuple)

    magnification = Setting(float(), schema_only=True)
    pixelSize = Setting(str(), schema_only=True)
    kind = Setting(bool(), schema_only=True)
    rotate = Setting(bool(), schema_only=True)
    orientation = Setting(int(), schema_only=True)

    def __init__(self):
        super().__init__()
        qt.QLocale.setDefault(qt.QLocale("en_US"))
        self._widget = None
        # Add combobox to choose between RSM and magnification (in one dimension case)
        self._methodCB = qt.QComboBox(self)
        self._methodCB.addItems(["RSM", "Magnification"])
        self._methodCB.hide()
        self.controlArea.layout().addWidget(self._methodCB)
        self._methodCB.currentTextChanged.connect(self._changeWidget)

    @Inputs.dataset
    def setDataset(self, _input):
        if _input is not None:
            if self._widget:
                self.controlArea.layout().removeWidget(self._widget)
                self._widget.hide()
            dataset, update = _input
            self._dataset = dataset
            if not dataset[1].dims.ndim:
                msg = qt.QMessageBox()
                msg.setIcon(qt.QMessageBox.Warning)
                msg.setText(
                    "This widget has to be used before selecting any region of \
                             interest and after selecting the dimensions"
                )
                msg.exec_()
            else:
                if dataset[1].dims.ndim == 1:
                    self._methodCB.show()
                    self._changeWidget("RSM")
                else:
                    self._methodCB.hide()
                    self._widget = MagnificationWidget(parent=self)
                    if self.magnification:
                        self._widget.magnification = self.magnification
                        self._widget.orientation = self.orientation

                    self._widget.sigComputed.connect(self._sendSignal)
                    self.controlArea.layout().addWidget(self._widget)
                    self._widget.setDataset(*dataset)

            if update is None:
                self.open()
            elif update != self._widget:
                self.Outputs.dataset.send(((self,) + self._widget.getDataset(), update))

    def _updateDataset(self, widget, dataset):
        self._widget._updateDataset(widget, dataset)
        if widget != self:
            self.Outputs.dataset.send(((self,) + self._widget.getDataset(), widget))

    def _changeWidget(self, method):
        """
        Change the widget displayed on the window
        """
        if self._widget:
            self.controlArea.layout().removeWidget(self._widget)
            self._widget.hide()
        if method == "RSM":
            self._widget = RSMWidget(parent=self)
            if self.pixelSize:
                self._widget.pixelSize = self.pixelSize
                self._widget.rotate = self.rotate
        else:
            self._widget = MagnificationWidget(parent=self)
            if self.magnification:
                self._widget.magnification = self.magnification
                self._widget.orientation = self.orientation
        self._widget.sigComputed.connect(self._sendSignal)
        self.controlArea.layout().addWidget(self._widget)
        self._widget.setDataset(*self._dataset)

    def _sendSignal(self):
        """
        Emits the signal with the new dataset.
        """
        if hasattr(self._widget, "magnification"):
            self.magnification = self._widget.magnification
            self.kind = False
            self.orientation = self._widget.orientation
        elif hasattr(self._widget, "pixelSize"):
            self.pixelSize = self._widget.pixelSize
            self.rotate = self._widget.rotate
            self.kind = True
        self.close()
        self.Outputs.dataset.send(((self,) + self._widget.getDataset(), None))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "29/09/2021"

import numpy

try:
    import tqdm
except ImportError:
    tqdm = None
from functools import partial
import multiprocessing
from multiprocessing import Pool
from scipy.optimize import curve_fit
from scipy.stats import skew, kurtosis
from skimage.transform import rescale
from silx.math.medianfilter import medfilt2d

import darfix


def gaussian(x, a, b, c, d):
    """
    Function to calculate the Gaussian with constants a, b, and c

    :param float x: Value to evaluate
    :param float a: height of the curve's peak
    :param float b: position of the center of the peak
    :param float c: standard deviation
    :param float d: lowest value of the curve (value of the limits)

    :returns: result of the function on x
    :rtype: float
    """
    return d + a * numpy.exp(-numpy.power(x - b, 2) / (2 * numpy.power(c, 2)))


def multi_gaussian(M, x0, y0, xalpha, yalpha, A, C, bg):
    """
    Bivariate case of the multigaussian PDF + background
    """
    x, y = M
    return bg + A * numpy.exp(
        -0.5
        / (1 - C**2)
        * (
            ((x - x0) / xalpha) ** 2
            + ((y - y0) / yalpha) ** 2
            - 2 * C * (x - x0) * (y - y0) / xalpha / yalpha
        )
    )


def generator(data, moments=None, indices=None):
    """
    Generator that returns the rocking curve for every pixel

    :param ndarray data: data to analyse
    :param moments: array of same shape as data with the moments values per pixel and image, optional
    :type moments: Union[None, ndarray]
    """
    for i in range(data.shape[1]):
        for j in range(data.shape[2]):
            if indices is None:
                new_data = data[:, i, j]
            else:
                new_data = numpy.zeros(data.shape[0])
                new_data[indices] = data[indices, i, j]
            if moments is not None:
                yield new_data, moments[:, i, j]
            yield new_data, None


def generator_2d(data, moments=None):
    """
    Generator that returns the rocking curve for every pixel

    :param ndarray data: data to analyse
    :param moments: array of same shape as data with the moments values per pixel and image, optional
    :type moments: Union[None, ndarray]
    """
    for i in range(data.shape[2]):
        for j in range(data.shape[3]):
            yield data[:, :, i, j], None


def fit_rocking_curve(y_values, values=None, num_points=None, int_thresh=None):
    """
    Fit rocking curve.

    :param tuple y_values: the first element is the dependent data and the second element are
        the moments to use as starting values for the fit
    :param values: The independent variable where the data is measured, optional
    :type values: Union[None, list]
    :param int num_points: Number of points to evaluate the data on, optional
    :param float int_thresh: Intensity threshold. If not None, only the rocking curves with
        higher ptp (range of values) are fitted, others are assumed to be noise and not important
        data. This parameter is used to accelerate the fit. Optional.

    :returns: If curve was fitted, the fitted curve, else item[0]
    :rtype: list
    """
    y, moments = y_values
    y = numpy.asanyarray(y)
    x = numpy.asanyarray(values) if values is not None else numpy.arange(len(y))
    if int_thresh is not None and y.ptp() < int_thresh:
        return y, [0, x[0], 0, min(y)]
    if moments is not None:
        p0 = [y.ptp(), moments[0], moments[1], min(y)]
    else:
        _sum = sum(y)
        if _sum > 0:
            mean = sum(x * y) / sum(y)
            sigma = numpy.sqrt(sum(y * (x - mean) ** 2) / sum(y))
        else:
            mean, sigma = numpy.nan, numpy.nan
        p0 = [y.ptp(), mean, sigma, min(y)]
    if numpy.isnan(mean) or numpy.isnan(sigma):
        return y, p0
    if numpy.isclose(p0[2], 0):
        return y, p0
    if num_points is None:
        num_points = len(y)
    epsilon = 1e-2
    bounds = numpy.array(
        [
            [min(y.ptp(), min(y)) - epsilon, min(x) - epsilon, 0, -numpy.inf],
            [max(max(y), y.ptp()) + epsilon, max(x) + epsilon, numpy.inf, numpy.inf],
        ]
    )

    p0 = numpy.array(p0)
    p0[p0 < bounds[0]] = bounds[0][p0 < bounds[0]]
    p0[p0 > bounds[1]] = bounds[1][p0 > bounds[1]]
    try:
        pars, cov = curve_fit(f=gaussian, xdata=x, ydata=y, p0=p0, bounds=bounds)
        y_gauss = gaussian(numpy.linspace(x[0], x[-1], num_points), *pars)
        y_gauss[numpy.isnan(y_gauss)] = 0
        y_gauss[y_gauss < 0] = 0
        pars[2] *= darfix.config.FWHM_VAL
        return y_gauss, pars
    except RuntimeError:
        p0[2] *= darfix.config.FWHM_VAL
        return y, p0
    except ValueError:
        p0[2] *= darfix.config.FWHM_VAL
        return y, p0


def fit_2d_rocking_curve(y_values, values, shape, int_thresh=None):

    y, moments = y_values
    y = numpy.asanyarray(y)
    A = y.ptp()
    values = numpy.asanyarray(values)
    _sum = sum(y)
    if numpy.isclose(_sum, 0, rtol=1e-03):
        return y, [numpy.nan, numpy.nan, numpy.nan, numpy.nan, A, 0, 0]
    x0 = sum(values[0] * y) / _sum
    y0 = sum(values[1] * y) / _sum
    xalpha = numpy.sqrt(sum(y * (values[0] - x0) ** 2) / _sum)
    yalpha = numpy.sqrt(sum(y * (values[1] - y0) ** 2) / _sum)
    if (int_thresh is not None and y.ptp() < int_thresh) or xalpha == 0 or yalpha == 0:
        return y, [x0, y0, xalpha, yalpha, A, 0, 0]
    X, Y = numpy.meshgrid(
        values[0, : shape[0]], values[1].reshape(numpy.flip(shape))[:, 0]
    )
    xdata = numpy.vstack((X.ravel(), Y.ravel()))
    epsilon = 1e-3
    try:
        pars, cov = curve_fit(
            f=multi_gaussian,
            xdata=xdata,
            ydata=y,
            p0=[x0, y0, xalpha, yalpha, A, 0, 0],
            bounds=(
                [
                    min(values[0]) - epsilon,
                    min(values[1]) - epsilon,
                    -numpy.inf,
                    -numpy.inf,
                    min(y.ptp(), min(y)) - epsilon,
                    -1,
                    -numpy.inf,
                ],
                [
                    max(values[0]) + epsilon,
                    max(values[1]) + epsilon,
                    numpy.inf,
                    numpy.inf,
                    max(y.ptp(), max(y)) + epsilon,
                    1,
                    numpy.inf,
                ],
            ),
        )
        y_gauss = multi_gaussian([X, Y], *pars)
        pars[2] *= darfix.config.FWHM_VAL
        pars[3] *= darfix.config.FWHM_VAL
        return y_gauss.ravel(), pars
    except RuntimeError:
        return y, [
            x0,
            y0,
            darfix.config.FWHM_VAL * xalpha,
            darfix.config.FWHM_VAL * yalpha,
            A,
            0,
            0,
        ]


def fit_data(
    data,
    moments=None,
    values=None,
    shape=None,
    indices=None,
    int_thresh=15,
    _tqdm=False,
):
    """
    Fit data in axis 0 of data

    :param bool _tqdm: If True, execut fitting under tqdm library.
    :returns: fitted data
    """

    g = generator(data, moments)
    cpus = multiprocessing.cpu_count()
    curves, maps = [], []
    with Pool(cpus - 1) as p:
        if _tqdm and tqdm is not None:
            for curve, pars in tqdm.tqdm(
                p.imap(
                    partial(fit_rocking_curve, values=values, int_thresh=int_thresh), g
                ),
                total=data.shape[1] * data.shape[2],
            ):
                curves.append(list(curve))
                maps.append(list(pars))
        else:
            for curve, pars in p.map(
                partial(fit_rocking_curve, values=values, int_thresh=int_thresh), g
            ):
                curves.append(list(curve))
                maps.append(list(pars))

    return numpy.array(curves).T.reshape(data.shape), numpy.array(maps).T.reshape(
        (4, data.shape[-2], data.shape[-1])
    )


def fit_2d_data(
    data, values, shape, moments=None, int_thresh=15, indices=None, _tqdm=False
):
    """
    Fit data in axis 0 of data

    :param bool _tqdm: If True, execut fitting under tqdm library.
    :returns: fitted data
    """
    g = generator(data, moments, indices)
    cpus = multiprocessing.cpu_count()
    curves, maps = [], []
    with Pool(cpus - 1) as p:
        if _tqdm and tqdm is not None:
            for curve, pars in tqdm.tqdm(
                p.imap(
                    partial(
                        fit_2d_rocking_curve,
                        values=values,
                        shape=shape,
                        int_thresh=int_thresh,
                    ),
                    g,
                ),
                total=data.shape[-2] * data.shape[-1],
            ):
                curves.append(list(curve))
                maps.append(list(pars))
        else:
            for curve, pars in p.map(
                partial(
                    fit_2d_rocking_curve,
                    values=values,
                    shape=shape,
                    int_thresh=int_thresh,
                ),
                g,
            ):
                curves.append(list(curve))
                maps.append(list(pars))
    curves = numpy.array(curves).T
    if indices is not None:
        curves = curves[indices]
    return curves.reshape(data[indices].shape), numpy.array(maps).T.reshape(
        (7, data.shape[-2], data.shape[-1])
    )


def compute_moments(values, data):
    """
    Compute first, second, third and fourth moment of data on values.

    :returns: The four moments
    """
    stack = [values[i] * data[i] for i in range(len(data))]
    zsum = numpy.sum(data, axis=0)
    com = numpy.sum(stack, axis=0) / zsum
    com = medfilt2d(com.astype(numpy.float64))
    repeat_values = numpy.repeat(values, com.shape[0] * com.shape[1]).reshape(
        len(values), com.shape[0], com.shape[1]
    )
    std = (
        numpy.sqrt(
            sum((data[i] * ((repeat_values[i] - com) ** 2) for i in range(len(data))))
        )
        / zsum
    )
    std = medfilt2d(std.astype(numpy.float64))
    skews = skew(stack, axis=0)
    kurt = kurtosis(stack, axis=0)

    return com, darfix.config.FWHM_VAL * std, skews, kurt


def compute_peak_position(data, values=None, center_data=False):
    """
    Compute peak position map

    :param bool center_data: If True, the values are centered on 0.
    """
    if values is not None:
        values = numpy.asanyarray(values)
        x = numpy.array(numpy.argmax(data, axis=0))
        if center_data:
            middle = float(min(values) + values.ptp()) / 2
            values -= middle
        image = [values[i] for i in x.flatten()]
        image = numpy.reshape(image, x.shape)
    else:
        image = numpy.array(numpy.argmax(data, axis=0))
        if center_data:
            middle = len(data) / 2
            vals = numpy.linspace(-middle, middle, len(data))
            image = image * numpy.ptp(vals) / len(data) + numpy.min(vals)
    return image


def compute_rsm(H, W, d, ffz, mainx, rotate=False):
    """
    Transformation to azimuthal coordinates.

    :param int H: height of the image in pixels.
    :param int W: width of the image in pixels.
    :param float d: Distance in micrometers of each pixel.
    :param float ffz: motor 'ffz' value.
    :param float mainx: motor 'mainx' value.

    :returns: Tuple of two arrays of size (W, H)
    :rtype: (X1, X2) : ndarray
    """
    if rotate:
        y = (numpy.arange(H) - H / 2) * d
        z = ffz - (W / 2 - numpy.arange(W)) * d
        y, z = numpy.meshgrid(y, z, indexing="ij")
    else:
        y = (numpy.arange(W) - W / 2) * d
        z = ffz - (H / 2 - numpy.arange(H)) * d
        z, y = numpy.meshgrid(z, y, indexing="ij")
    eta = numpy.arctan2(y, z)
    twotheta = numpy.arctan2(numpy.sqrt(y * y + z * z), mainx)
    return numpy.degrees(eta), numpy.degrees(twotheta)


def compute_magnification(
    H, W, d, obx, obpitch, mainx, topography=[False, 0], center=True
):
    """
    :param int H: height of the image in pixels.
    :param int W: width of the image in pixels.
    :param float d: Distance in micrometers of each pixel.
    :param float obx: motor 'obx' value.
    :param float obpitch: motor 'obpitch' value in the middle of the dataset.
    :param float mainx: motor 'mainx' value.

    :returns: Tuple of two arrays of size (H, W)
    :rtype: (X1, X2) : ndarray
    """

    pix_arr = numpy.meshgrid(numpy.arange(H), numpy.arange(W))
    d1 = obx / numpy.cos(numpy.radians(obpitch))
    d2 = mainx / numpy.cos(numpy.radians(obpitch)) - d1
    M = d2 / d1
    d /= M
    if center:
        pix_arr[0] = (pix_arr[0] - W / 2) * d
        pix_arr[1] = (H / 2 - pix_arr[1]) * d
    else:
        pix_arr[0] = pix_arr[0] * d
        pix_arr[1] = (H - 1 - pix_arr[1]) * d
    if topography[0]:
        pix_arr[topography[1]] /= numpy.sin(numpy.radians(obpitch))
    return pix_arr


def rescale_data(data, scale):
    new_data = []
    for image in data:
        new_data += [rescale(image, scale, anti_aliasing=True, preserve_range=True)]
    return numpy.array(new_data)


def calculate_RSM_histogram(
    data,
    diffry_values,
    twotheta,
    eta,
    Q,
    a,
    map_range,
    units="poulsen",
    map_shape=(50, 50, 50),
    n=[1, 0, 0],
    E=17,
):
    """
    ***Code originally written by Mads Carslen***

    Calculate reciprocal space map from a 'diffry' scan without the objective lens.
    The RSM is calculated as a multidimensional histogram;
    :param array_like scan: usual scan object
    :param tuple ROI: tuple containing two slice objects to give the ROI.
    :param Union[list, tuple] Q: Scattering vector in oriented pseudocubic coordinates.
    :param float a: pseudocubic lattice parameter
    :param float map_range: range (in all 3 directions) of the histogram. Center-to edge-distance.
    :param str detector: either 'basler' or 'pco' depending on which detector is used. 'pco' implementation is not well-tested at the moment
    :param str units: either 'poulsen'  [10.1107/S1600576717011037] or 'gorfman' [https://arxiv.org/pdf/2110.14311.pdf]
    :param float diffry_center: origin of the diffry axis. This changes the orientation of the sampe reciprocal space relative to lab frame
    :param array_like detect_orig: pixel coordinate of the detector origin, which is assumed to be placed at (x_detec, 0, ffz) relative to the sample
    :param float x_detec: x position of the detector
    :param array_like map_shape: Number of bins in each direction
    :param array_like n: surface normal of the sample in oriented pseudocubic hkl
    :param float E: energy
    """

    k = 2 * numpy.pi / (12.391 / E)

    diffry_center = numpy.mean(diffry_values)
    img_shape = data[0].shape
    if units == "gorfman":
        # Build orientation matrix
        sampl_z = numpy.array(Q) / numpy.linalg.norm(
            numpy.array(Q)
        )  # assume scattering vector is z
        sampl_x = n - sampl_z * numpy.dot(sampl_z, n) / numpy.linalg.norm(
            n
        )  # orthogonalize
        sampl_x /= numpy.linalg.norm(sampl_x)  # nomalize

        sampl_y = numpy.cross(sampl_z, sampl_x)
        lab_to_lat = numpy.stack((sampl_x, sampl_y, sampl_z))
    elif units == "poulsen":
        lab_to_lat = numpy.identity(3)

    # Calculate lab frame q vector for each pixel
    k0 = numpy.array([k, 0, 0])  # Lab frame incidetn wavevector
    kh = k * numpy.stack(
        [
            numpy.cos(twotheta),
            numpy.sin(twotheta) * numpy.sin(eta),
            numpy.sin(twotheta) * numpy.cos(eta),
        ]
    )  # Lab frame scattered wavevector
    q = kh - k0[:, numpy.newaxis, numpy.newaxis]
    if units == "gorfman":
        q = q * a / 2 / numpy.pi
    elif units == "poulsen":
        q = q * a / 2 / numpy.pi / numpy.linalg.norm(Q)

    # flatten to match syntax for histogramdd
    q = q.reshape(3, img_shape[0] * img_shape[1])
    # Rotate from lab frame to sample frame
    theta_ref = numpy.arcsin(2 * numpy.pi * numpy.linalg.norm(Q) / a / k / 2)
    q = numpy.stack(
        [
            q[0, ...] * numpy.cos(theta_ref) + q[2, ...] * numpy.sin(theta_ref),
            q[1, ...],
            q[2, ...] * numpy.cos(theta_ref) - q[0, ...] * numpy.sin(theta_ref),
        ]
    )

    # Make histogram ranges
    q_mean = numpy.mean(q, axis=1)
    diffry_mean = (numpy.mean(diffry_values) - diffry_center) * numpy.pi / 180
    q_mean = numpy.stack(
        [
            q_mean[0] * numpy.cos(diffry_mean) - q_mean[2] * numpy.sin(diffry_mean),
            q_mean[1],
            q_mean[2] * numpy.cos(diffry_mean) + q_mean[0] * numpy.sin(diffry_mean),
        ]
    )

    if units == "gorfman":
        q_mean = lab_to_lat.transpose() @ q_mean

    ranges = (
        (q_mean[0] - map_range, q_mean[0] + map_range),
        (q_mean[1] - map_range, q_mean[1] + map_range),
        (q_mean[2] - map_range, q_mean[2] + map_range),
    )  # hkl units

    # initialize sum arrays
    sum_inte = numpy.zeros(map_shape)
    sum_freq = numpy.zeros(map_shape)

    if tqdm is not None:
        loop = tqdm.tqdm(data)
    else:
        loop = data
    # Loop through images
    for i, image in enumerate(loop):
        # read angle
        diffry = (diffry_values[i] - diffry_center) * numpy.pi / 180
        # rotate q back to zero-motor frame
        q_rot = numpy.stack(
            [
                q[0, ...] * numpy.cos(diffry) - q[2, ...] * numpy.sin(diffry),
                q[1, ...],
                q[2, ...] * numpy.cos(diffry) + q[0, ...] * numpy.sin(diffry),
            ]
        )
        # Rotate into lattice frame
        q_rot = lab_to_lat.transpose() @ q_rot
        # Do binning
        sum_inte += numpy.histogramdd(
            q_rot.transpose(), map_shape, ranges, weights=image.flatten()
        )[0]
        sum_freq += numpy.histogramdd(q_rot.transpose(), map_shape, ranges)[0]

    edges = numpy.histogramdd(q_rot.transpose(), map_shape, ranges)[1]
    arr = sum_inte / sum_freq
    arr[sum_freq == 0] = numpy.nan

    if units == "poulsen":
        edges[2] = edges[2] - 1

    return arr, edges



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "28/04/2020"

import numpy
import unittest

from sklearn.datasets import load_iris
from sklearn.decomposition import PCA

from darfix.decomposition.ipca import IPCA
from darfix.decomposition.nica import NICA
from darfix.decomposition.test.utils import images, sampler


class TestNICA(unittest.TestCase):
    """Tests for `nica.py`."""

    def setUp(self):
        iris = load_iris()
        self.X = iris.data

    def test_whiten_in_memory(self):
        nica = NICA(self.X, 2)
        pca = PCA(2, whiten=True)
        X = pca.fit_transform(self.X.T).T
        numpy.testing.assert_equal(X, nica.Z)
        numpy.testing.assert_equal(pca.components_, nica.V)

    def test_whiten_in_disk(self):
        nica = NICA(self.X, 2, chunksize=2)
        ipca = IPCA(self.X, 2, 2, whiten=True, rowvar=False)
        ipca.fit_transform()
        numpy.testing.assert_equal(ipca.W.T, nica.Z)
        numpy.testing.assert_equal(ipca.H, nica.V)

    def test_W(self):
        nica = NICA(self.X, 2)
        nica.fit_transform()
        self.assertEqual(nica.W.shape, (self.X.shape[0], 2))

    def test_H(self):
        nica = NICA(self.X, 2)
        nica.fit_transform()
        self.assertEqual(nica.H.shape, (2, self.X.shape[1]))

    def test_fit_transform(self):
        resources = ["circle", "star", "pentagon", "square"]
        num_images = 100
        means = [15, 30, 45, 60]
        tol = 0.7
        sample = sampler(resources, means)

        X = numpy.array([sample(i).flatten() for i in range(num_images)])
        nica = NICA(X, 4)
        nica.fit_transform(max_iter=1000)

        stack = numpy.asarray(images(resources))
        for img in stack:
            img = img / numpy.linalg.norm(img)
            found = False
            for component in nica.H:
                comp = component.reshape(img.shape)
                comp = comp / numpy.linalg.norm(comp)
                err = numpy.linalg.norm(comp - img)
                if err < tol:
                    found = True
                    break
            self.assertTrue(found)

    def test_fit_transform_IPCA(self):
        resources = ["circle", "star", "pentagon", "square"]
        num_images = 100
        means = [15, 30, 45, 60]
        tol = 0.7
        sample = sampler(resources, means)

        X = numpy.array([sample(i).flatten() for i in range(num_images)])
        nica = NICA(X, 4, chunksize=1000)
        nica.fit_transform(max_iter=1000)

        stack = numpy.asarray(images(resources))
        for img in stack:
            img = img / numpy.linalg.norm(img)
            found = False
            for component in nica.H:
                comp = component.reshape(img.shape)
                comp = comp / numpy.linalg.norm(comp)
                err = numpy.linalg.norm(comp - img)
                if err < tol:
                    found = True
                    break
            self.assertTrue(found)

    def test_fit_transform_indices(self):
        resources = ["circle", "star", "pentagon", "square"]
        num_images = 100
        means = [15, 30, 45, 60]
        tol = 0.7
        sample = sampler(resources, means)

        X = numpy.array([sample(i).flatten() for i in range(num_images)])
        nica = NICA(X, 4, indices=numpy.arange(50))
        nica.fit_transform(max_iter=1000)

        stack = numpy.asarray(images(resources[:-1]))
        for img in stack:
            img = img / numpy.linalg.norm(img)
            found = False
            for component in nica.H:
                comp = component.reshape(img.shape)
                comp = comp / numpy.linalg.norm(comp)
                err = numpy.linalg.norm(comp - img)
                if err < tol:
                    found = True
                    break
            self.assertTrue(found)

    def test_fit_transform_IPCA_indices(self):
        resources = ["circle", "star", "pentagon", "square"]
        num_images = 100
        means = [15, 30, 45, 60]
        tol = 0.7
        sample = sampler(resources, means)

        X = numpy.array([sample(i).flatten() for i in range(num_images)])
        nica = NICA(X, 4, chunksize=1000, indices=numpy.arange(50))
        nica.fit_transform(max_iter=1000)

        stack = numpy.asarray(images(resources[:-1]))
        for img in stack:
            img = img / numpy.linalg.norm(img)
            found = False
            for component in nica.H:
                comp = component.reshape(img.shape)
                comp = comp / numpy.linalg.norm(comp)
                err = numpy.linalg.norm(comp - img)
                if err < tol:
                    found = True
                    break
            self.assertTrue(found)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "03/12/2020"


from silx.gui import qt


class ChooseDimensionDock(qt.QDockWidget):
    def __init__(self, parent=None, vertical=True, values=True, _filter=True):
        """
        Dock widget containing the ChooseDimensionWidget.
        """
        qt.QDockWidget.__init__(self, parent)
        self.widget = ChooseDimensionWidget(self, vertical, values, _filter)
        self.setWidget(self.widget)


class ChooseDimensionWidget(qt.QWidget):
    """
    Widget to choose a dimension from a dict and choose the value to filter
    the data. It can be included in other widget like StackView to filter the
    stack.
    """

    filterChanged = qt.Signal(list, list)
    stateDisabled = qt.Signal()

    def __init__(self, parent=None, vertical=True, values=True, _filter=True):
        qt.QWidget.__init__(self, parent)

        if vertical:
            self.setLayout(qt.QVBoxLayout())
        else:
            self.setLayout(qt.QHBoxLayout())
        self.show_values = values
        self.dimensionWidgets = []
        self.filter = _filter
        if _filter:
            self._checkbox = qt.QCheckBox("Filter by dimension", self)
            self._checkbox.stateChanged.connect(self._updateState)

    def setDimensions(self, dimensions):
        """
        Function that fills the corresponeding comboboxes with the dimension's
        name and possible values.

        :param array_like dimensions: List of `darfix.core.dataset.Dimension`
                                      elements.
        """
        self.dimensionWidgets = []
        for i in reversed(range(self.layout().count())):
            self.layout().itemAt(i).widget().setParent(None)
        self.dimensions = dimensions
        self.dimension = []
        self.value = [0 for i in range(self.dimensions.ndim - 1)]
        for i in range(dimensions.ndim - 1):
            self._addDimensionWidget()
            self.dimension.append(i)
            self.dimensionWidgets[-1][0].setCurrentIndex(i)

        status = self.blockSignals(True)
        self._updateDimension(0)
        if self.filter:
            self.blockSignals(status)
            self._updateState(self._checkbox.isChecked())
            self.layout().addWidget(self._checkbox)

    def _addDimensionWidget(self):
        """
        Add new widget to choose between different dimensions and values.
        """

        widget = qt.QWidget(self)
        layout = qt.QGridLayout()
        dimensionCB = qt.QComboBox()
        valueCB = None
        if self.show_values:
            dimensionLabel = qt.QLabel("Dimension: ")
            layout.addWidget(dimensionLabel, 0, 0)
            layout.addWidget(dimensionCB, 0, 1)
            valueLabel = qt.QLabel("Value: ")
            valueCB = qt.QComboBox()
            layout.addWidget(valueLabel, 1, 0)
            layout.addWidget(valueCB, 1, 1)
            valueCB.setEnabled(False)
            valueCB.currentIndexChanged.connect(self._updateValue)
        else:
            layout.addWidget(dimensionCB, 0, 0)
        self.dimensionWidgets.append([dimensionCB, valueCB])
        widget.setLayout(layout)
        self.layout().addWidget(widget)
        dimensionCB.setEnabled(False)

        for dimension in self.dimensions:
            dimensionCB.insertItem(dimension[0], dimension[1].name)
        dimensionCB.currentIndexChanged.connect(self._updateDimension)

    def _updateDimension(self, axis=-1):
        """
        Updates the selected dimension and set's the corresponding possible values.

        :param int axis: selected dimension's axis, only used to check valid call
            of the method.
        """
        if axis != -1 and axis is not None:
            self.dimension = []
            # Init values to 0
            self.value = [0 for i in range(self.dimensions.ndim - 1)]
            # Reset all dimensions
            for dimWidget in self.dimensionWidgets:
                # Prevent signals
                status = dimWidget[0].blockSignals(True)
                axis = dimWidget[0].currentIndex()
                # Enable / disable items in combobox
                for dimension in self.dimensions:
                    if dimension[0] in self.dimension:
                        dimWidget[0].model().item(dimension[0]).setEnabled(False)
                        # If axis is already in the dimensions list, set it to
                        # next available axis.
                        if axis == dimension[0]:
                            axis = (axis + 1) % self.dimensions.ndim
                            dimWidget[0].setCurrentIndex(axis)
                    else:
                        dimWidget[0].model().item(dimension[0]).setEnabled(True)
                dimWidget[0].blockSignals(status)
                if dimWidget[0].currentText() != "None":
                    self.dimension.append(axis)
                if self.show_values:
                    # Set values from new dimension
                    dimWidget[1].clear()
                    dimWidget[1].addItems(
                        map(str, self.dimensions.get(axis).unique_values)
                    )

            self.filterChanged.emit(self.dimension, self.value)

    def _updateValue(self, index=None):
        """
        Updates the selected value.

        :param int index: selected value's index.
        """
        if self.show_values and (index is not None or index != -1):
            self.value = []
            for dimWidget in self.dimensionWidgets:
                axis = dimWidget[1].currentIndex()
                self.value.append(axis)

            self.filterChanged.emit(self.dimension, self.value)

    def _updateState(self, state):
        """
        Updates the state of the widget.

        :param bool state: If True, the widget emit's a signal
                    with the selected dimension and value. Else,
                    a disabled signal is emitted.

        """
        for dimWidget in self.dimensionWidgets:
            dimWidget[0].setEnabled(state)
            if self.show_values:
                dimWidget[1].setEnabled(state)

        if state:
            self.filterChanged.emit(self.dimension, self.value)
        else:
            self.stateDisabled.emit()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno", "J. Garriga"]
__license__ = "MIT"
__date__ = "26/04/2021"

from functools import partial

from silx.gui import qt
from darfix.core.dataset import Dataset
from darfix.core.dimension import (
    Dimension,
    _METADATA_TYPES,
    _METADATA_TYPES_I,
    DEFAULT_METADATA,
)
import numpy
import logging

_logger = logging.getLogger(__file__)


class DimensionMapping(qt.QWidget):
    """
    Widget used to define the number of dimension and with which values they are
    mapped
    """

    _V_HEADERS = ["Axis", "Kind", "Name", "Size", "Range", "Tolerance", "", ""]

    fitSucceed = qt.Signal()
    """Signal emitted when the fit succeeds"""
    fitFailed = qt.Signal()
    """Signal emitted when the fit fails"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        self._table = qt.QTableWidget(parent=self)
        self._table.setColumnCount(len(self._V_HEADERS))
        self._table.setHorizontalHeaderLabels(self._V_HEADERS)
        header = self._table.horizontalHeader()
        header.setMinimumSectionSize(50)
        if qt.qVersion() < "5.0":
            setResizeMode = header.setResizeMode
        else:
            setResizeMode = header.setSectionResizeMode
        for iColumn in range(len(self._V_HEADERS)):
            if iColumn in (1, 2):
                setResizeMode(iColumn, qt.QHeaderView.Stretch)
            else:
                setResizeMode(iColumn, qt.QHeaderView.ResizeToContents)
        self._table.verticalHeader().hide()
        self._dims = {}

        self.layout().addWidget(self._table, 0, 0, 6, 10)
        self._addButton = qt.QPushButton("Add", parent=self)
        self.layout().addWidget(self._addButton, 6, 6, 1, 1)

        # connect Signal/SLOT
        self._addButton.pressed.connect(self.addDim)

    def clear(self):
        widgets = list(self._dims.values())
        for widget in widgets:
            self.removeDim(widget)

    @property
    def ndim(self):
        return len(self._dims)

    @property
    def dims(self):
        return self._dims

    def addDim(self, axis=None, dim=None):
        """

        :param axis: which axis is defining this dimension
        :param `Dimension` dim: definition of the dimension to add
        """
        if axis is None:
            axis = self._getNextFreeAxis()
        row = self._table.rowCount()
        self._table.setRowCount(row + 1)
        widget = _DimensionItem(parent=self, table=self._table, row=row)
        widget.removed.connect(self.removeDim)
        if dim is not None:
            widget.setDim(dim)
        widget.axis = axis
        self._dims[row] = widget
        return widget

    def removeDim(self, row):
        """
        Remove dimension.

        :param Union[int,`_DimensionItem`]: row or item to remove
        """
        if isinstance(row, _DimensionItem):
            iRow = row._row
        else:
            iRow = row
        self._table.removeRow(iRow)
        self._dims[iRow].removed.disconnect(self.removeDim)
        self._dims[iRow].setAttribute(qt.Qt.WA_DeleteOnClose)
        self._dims[iRow].close()
        del self._dims[iRow]
        ini_rows = sorted(list(self._dims.keys()))
        for row in ini_rows:
            if row <= iRow:
                continue
            widget = self._dims[row]
            new_row = row - 1
            assert new_row >= 0
            widget.embedInTable(table=self._table, row=new_row)
            widget.axis = self._getNextFreeAxis()
            self._dims[new_row] = widget
            del self._dims[row]

        self.fitFailed.emit()

    def _getNextFreeAxis(self):
        """
        :return int: next unused axis
        """
        res = 0
        usedAxis = []
        [usedAxis.append(_dim.axis) for _dim in self._dims.values()]
        while res in usedAxis:
            res = res + 1
        return res


class DimensionWidget(DimensionMapping):
    """
    Widget to define dimensions and try to fit those with dataset
    """

    def __init__(self, parent=None):
        DimensionMapping.__init__(self, parent)
        self._dataset = None
        self._update_dataset = None
        self.indices = None
        self.bg_indices = None
        self.bg_dataset = None
        self.metadata_type = _METADATA_TYPES["positioner"]
        self.tolerance = 1e-9

        metadataTypeLabel = qt.QLabel("Metadata type: ")
        toleranceLabel = qt.QLabel("Tolerance: ")
        self._metadataTypeCB = qt.QComboBox()
        for metaType in _METADATA_TYPES:
            self._metadataTypeCB.addItem(metaType)
        self._toleranceLE = qt.QLineEdit(str(self.tolerance))
        self._toleranceLE.setToolTip(
            "Tolerance for which the values in the dimensions will be considered unique"
        )
        validator = qt.QDoubleValidator()
        validator.setBottom(0)
        self._toleranceLE.setValidator(validator)
        self._autoDetect = qt.QPushButton("Find dimensions", parent=self)
        self._autoDetect.setToolTip(
            "Automatically find all dimensions that "
            "change through the dataset. \n"
            "It is considered that a dimension changes "
            "if the number of unique values is greater "
            "than one.\nThe metadata type is needed to "
            "choose which values to compare, and the "
            "threshold to know when two values are "
            "considered to be different.\nThe threshold"
            " is only used in values that are numbers."
        )
        self._fitButton = qt.QPushButton("Fit", parent=self)
        self._fitButton.setToolTip("Try to reshape data into given dimensions")
        self._clearButton = qt.QPushButton("Clear", parent=self)

        self.layout().addWidget(metadataTypeLabel, 6, 0, 1, 1)
        self.layout().addWidget(self._metadataTypeCB, 6, 1, 1, 1)
        self.layout().addWidget(toleranceLabel, 6, 2, 1, 1)
        self.layout().addWidget(self._toleranceLE, 6, 3, 1, 1)
        self.layout().addWidget(self._clearButton)
        self.layout().addWidget(self._fitButton)
        self.layout().addWidget(self._autoDetect, 6, 4, 1, 1)

        # connect Signal/SLOT
        self._fitButton.pressed.connect(self.fit)
        self._clearButton.pressed.connect(self.clear)
        self._autoDetect.pressed.connect(self._find_dimensions)
        self._metadataTypeCB.currentTextChanged.connect(self._updateKind)
        self._toleranceLE.textChanged.connect(self._updateTolerance)

    @property
    def dataset(self):
        return self._dataset

    def getDataset(self):
        return self._update_dataset, self.indices, self.bg_indices, self.bg_dataset

    def _updateKind(self, metadata_type=None):
        if metadata_type is None:
            metadata_type = self._metadataTypeCB.currentText()
        self.metadata_type = _METADATA_TYPES[metadata_type]

    def _updateTolerance(self, tolerance=None):
        if tolerance is None:
            tolerance = self._toleranceLE.text()
        try:
            self.tolerance = float(tolerance)
        except ValueError:
            _logger.warning("Wrong tolerance")

    def clear(self):
        super().clear()
        if self.dataset is None:
            _logger.warning("No dataset to be fitted")
            return
        self.dataset.clear_dims()

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """

        :param dataset: the dataset for which we want to define the
                           dimensions.
        :type dataset: `Dataset`
        """
        assert isinstance(dataset, Dataset)
        self._parent = parent
        self._dataset = dataset
        self._update_dataset = dataset
        self.indices = indices
        self.bg_indices = bg_indices
        self.bg_dataset = bg_dataset
        if dataset.nframes > 0:
            for widget in self._dims.values():
                widget._setMetadata(self.dataset.get_data().metadata[0])

    def _updateDataset(self, widget, dataset):
        del self._dataset
        self._dataset = dataset
        self._update_dataset = dataset
        self._parent._updateDataset(widget, dataset)

    def _showSnowMessage(self):
        msg = qt.QMessageBox()
        msg.setIcon(qt.QMessageBox.Critical)
        msg.setText("Dimension definition is not yet working with hdf5 files")
        msg.setWindowTitle("Don't eat yellow snow!")
        msg.setStandardButtons(qt.QMessageBox.Ok)
        msg.exec_()

    def _find_dimensions(self):
        """
        Automatically find all dimensions that change through the dataset.
        It is considered that a dimension changes if the number of unique values
        is greater than one. The metadata type is needed to choose which values
        to compare, and the threshold to know when two values are considered to
        be different. The threshold is only used in values that are numbers.
        """
        result = self.dataset.find_dimensions(self.metadata_type, self.tolerance)
        if not result:
            self._showSnowMessage()
        super().clear()
        for dim in self.dataset.dims:
            self.addDim(dim[0], dim[1])

    def fit(self):
        """
        Fit dimensions into the data.

        :returns: return status of the fit and fail reason, if any
        :rtype: Union[bool,str,None]
        """
        if self.dataset is None:
            _logger.warning("No dataset to be fitted")
            return

        qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
        msg = qt.QMessageBox()
        try:
            if self.dims:
                self.dataset.clear_dims()
                for dimension in self.dims.values():
                    axis = dimension.axis
                    dimension = dimension.dim
                    values = self.dataset.get_metadata_values(
                        kind=dimension.kind, key=dimension.name
                    )
                    indices = numpy.unique(values, return_index=True)[1]
                    dimension.set_unique_values(
                        [values[index] for index in sorted(indices)]
                    )
                    self.dataset.add_dim(axis=axis, dim=dimension)
                try:
                    self._update_dataset = self.dataset.reshape_data()
                    ndim = self._update_dataset.dims.ndim
                    for axis, dimension in self._update_dataset.dims:
                        # Flip axis to be consistent with the data shape
                        not_axis = ndim - axis - 1
                        if self._update_dataset.dims.ndim > 1:
                            metadata = numpy.swapaxes(
                                self._update_dataset.data.metadata, ndim - 1, not_axis
                            )
                            for axis in range(ndim - 1):
                                metadata = metadata[0]
                        else:
                            metadata = self._update_dataset.data.metadata
                        # missing_value = False
                        for data in metadata:
                            try:
                                values += [
                                    data.get_value(
                                        kind=dimension.kind, name=dimension.name
                                    )[0]
                                ]
                            except KeyError:
                                # missing_value = True
                                if len(values):
                                    values += [values[-1]]
                                else:
                                    values = ["0"]

                        dimension.set_unique_values(values)
                except ValueError:
                    for axis, dimension in self.dataset.dims:
                        values = numpy.unique(
                            self.dataset.get_metadata_values(
                                kind=dimension.kind, key=dimension.name
                            )
                        )
                        dimension.set_unique_values(values)
                    self._update_dataset = self.dataset.reshape_data()
            else:
                raise ValueError("Cannot fit without dimensions")
            super().clear()
            for dim in self.dataset.dims:
                self.addDim(dim[0], dim[1])
            self._update_dataset = self.dataset.reshape_data()
        except Exception as e:
            self.dataset.clear_dims()
            msg.setIcon(qt.QMessageBox.Warning)
            msg.setText("Error: {0}".format(e))
            msg.setWindowTitle("Fit failed!")
            self.fitFailed.emit()
            return False, e
        else:
            self.fitSucceed.emit()
            msg.setIcon(qt.QMessageBox.Information)
            msg.setText(
                "Dimensions {} could be fit".format(
                    ", ".join(self.dataset.dims.get_names())
                )
            )
            msg.setWindowTitle("Fit succeeded!")
            return True, None
        finally:
            qt.QApplication.restoreOverrideCursor()
            msg.setStandardButtons(qt.QMessageBox.Ok)
            msg.exec_()

    def setDims(self, dims):
        """

        :param dict dims: axis as key and `Dimension` as value.
        """
        self.clear()
        assert type(dims) is dict
        if self.dataset._isH5:
            self._showSnowMessage()
            return

        for axis, dim in dims.items():
            assert type(axis) is int
            assert isinstance(dim, Dimension)
            self.addDim(axis=axis, dim=dim)
            if self.dataset is not None and len(self.dataset.get_data().metadata) > 0:
                if not dim.unique_values:
                    values = self.dataset.get_metadata_values(dim.kind, dim.name)
                    dim.set_unique_values(values)
            self.dataset.add_dim(axis=axis, dim=dim)

    def addDim(self, axis=None, dim=None):
        """

        :param axis: which axis is defining this dimension
        :param `Dimension` dim: definition of the dimension to add
        """
        widget = super().addDim(axis, dim)
        if self.dataset is not None and len(self.dataset.data.metadata) > 0:
            widget._setMetadata(self.dataset.get_data().metadata[0])

        return widget


class _DimensionItem(Dimension, qt.QWidget):
    """Widget use to define a dimension"""

    removed = qt.Signal(qt.QObject)
    """Signal emitted when the Item should be removed"""

    dimValueChanged = qt.Signal()
    """Signal emitted when the dimension definition is changed"""

    axisChanged = qt.Signal(int, int)
    """Signal emitted when the axis value is changed: id (row), new_value_value
    """

    class _SizeWidget(qt.QWidget):
        valueChanged = qt.Signal(int)
        """Signal emitted when the value of the Spin box change but decorelated
        with the active state
        """

        def __init__(self, parent):
            qt.QWidget.__init__(self, parent)
            self.setLayout(qt.QVBoxLayout())
            self._sizeSP = qt.QSpinBox(parent=self)
            """_sizeSP will be used to define the size. Will be editable if the
            size editable (so if dimension is set from a relative parameter)
            """
            self.layout().addWidget(self._sizeSP)

            self.layout().setContentsMargins(0, 0, 0, 0)
            self._active = False
            self._sizeSP.hide()
            # expose API
            self.setMinimum = self._sizeSP.setMinimum

            # connect Signal/SLOT
            self._sizeSP.valueChanged.connect(self._valueHasChanged)

        @property
        def value(self):
            return self._sizeSP.value()

        def setValue(self, size, editable=True):
            """

            :param int or None size: the size of the dimension. If None, not
                                     define yet
            :param editable: the size will be editable if given by the use (so)
                             if is relative.
            """
            if size is None:
                self._sizeSP.hide()
                self._sizeLabel.hide()
            else:
                assert type(size) is int
                self._sizeSP.show()
                self._sizeSP.setMaximum(max(99, 2 * size))
                self._sizeSP.setValue(size)
                self.setEnabled(editable)

        def setEnabled(self, editable):
            self._sizeSP.setEnabled(editable)

        def toggle(self, checked):
            self._active = checked
            if checked:
                self._sizeSP.show()
            else:
                self._sizeSP.hide()

        def _valueHasChanged(self, value):
            self.valueChanged.emit(value)

    class _RangeWidget(qt.QWidget):
        def __init__(self, parent):
            qt.QWidget.__init__(self, parent)
            self.setLayout(qt.QHBoxLayout())
            self._minLE = qt.QLineEdit(parent=self)
            self._minLE.setPlaceholderText("Start")
            self._maxLE = qt.QLineEdit(parent=self)
            self._maxLE.setPlaceholderText("Stop")
            self._stepLE = qt.QLineEdit(parent=self)
            self._stepLE.setPlaceholderText("Step")
            self._minLE.setValidator(qt.QDoubleValidator())
            self._maxLE.setValidator(qt.QDoubleValidator())
            validator = qt.QDoubleValidator()
            validator.setBottom(0)
            self._stepLE.setValidator(validator)
            self.layout().addWidget(self._minLE)
            self.layout().addWidget(self._maxLE)
            self.layout().addWidget(self._stepLE)

            self.layout().setContentsMargins(0, 0, 0, 0)

        @property
        def value(self):
            if self._minLE.text() != "" and self._maxLE.text():
                step = float(self._stepLE.text()) if self._stepLE.text() != "" else 0
                return [float(self._minLE.text()), float(self._maxLE.text()), step]
            else:
                return None

        def setValue(self, _range):
            """

            :param int or None _range
            """
            assert type(_range) is list
            self._minLE.setText(str(numpy.round(_range[0], 5)))
            self._maxLE.setText(str(numpy.round(_range[1], 5)))
            self._stepLE.setText(str(numpy.round(_range[2], 5)))
            self._minLE.setCursorPosition(0)
            self._maxLE.setCursorPosition(0)
            self._stepLE.setCursorPosition(0)

    def __init__(self, parent, table, row):
        """

        :param QTableWidget table: if has to be embed in a table the
                                           parent table
        :param int row: row position in the QTableWidget. Also used as ID
        """
        qt.QWidget.__init__(self, parent)
        Dimension.__init__(self, kind=DEFAULT_METADATA, name="")

        self.__metadata = None

        # axis
        self._axis = qt.QSpinBox(parent=self)
        self._axis.setMinimum(0)
        # kind
        self._kindCB = qt.QComboBox(parent=self)
        self._kindCB.setMinimumWidth(100)
        for _kindName in _METADATA_TYPES:
            self._kindCB.addItem(_kindName)
        # name
        self._namesCB = qt.QComboBox(parent=self)
        self._namesCB.setMinimumWidth(100)
        # size
        self._sizeWidget = self._SizeWidget(parent=self)
        self._sizeWidget.setMinimum(0)
        self._setSize(0)
        # range
        self._rangeWidget = self._RangeWidget(parent=self)
        self._rangeWidget.setMinimumWidth(100)
        self._rangeWidget.setMaximumWidth(200)
        # tolerance
        self._toleranceLE = qt.QLineEdit(parent=self)
        self._toleranceLE.setMaximumWidth(80)
        validator = qt.QDoubleValidator()
        validator.setBottom(0)
        self._toleranceLE.setValidator(validator)
        self._setTolerance(1e-9)
        # rm button
        style = qt.QApplication.style()
        icon = style.standardIcon(qt.QStyle.SP_BrowserStop)
        self._rmButton = qt.QPushButton(icon=icon, parent=self)
        icon = style.standardIcon(qt.QStyle.SP_FileDialogContentsView)
        self._infoButton = qt.QPushButton(icon=icon, parent=self)
        self._infoButton.hide()

        # connect Signal/slot
        self._rmButton.pressed.connect(self.remove)
        self._axis.valueChanged.connect(self._axisHasChanged)
        self._kindCB.currentIndexChanged.connect(self._dimHasChanged)
        self._sizeWidget.valueChanged.connect(self._dimHasChanged)
        self._namesCB.currentIndexChanged.connect(self._dimHasChanged)
        self._kindCB.currentIndexChanged.connect(self._updateNames)
        self._infoButton.pressed.connect(self.showUniqueNames)

        _callback = partial(self._sizeWidget.setValue, 0)
        self._toleranceLE.editingFinished.connect(_callback)

        # update values from `Dim` class
        self._kindCB.currentTextChanged.connect(self.set_kind)
        self._namesCB.currentTextChanged.connect(self.set_name)
        self._sizeWidget.valueChanged.connect(self.set_size)

        self.embedInTable(table=table, row=row)
        self.__row = row

    def _axisHasChanged(self, value):
        self.axisChanged.emit(self._row, value)

    def _dimHasChanged(self, *args, **kwargs):
        self.dimValueChanged.emit()

    def remove(self):
        self.removed.emit(self)

    def showUniqueNames(self):
        title = "%s - %s" % (_METADATA_TYPES_I[self.kind], self.name)
        if len(self.unique_values):
            ptp = float(self.unique_values[-1]) - float(self.unique_values[0])
            unique_values_str = []
            [unique_values_str.append(str(val)) for val in self.unique_values]
            msg = (
                "Unique values: ["
                + ", ".join(unique_values_str)
                + "]\n Ptp: "
                + str(ptp)
            )
            qt.QMessageBox.information(self, title, msg)
        else:
            qt.QMessageBox.information(self, title, "No values were found, try fitting")

    def setUniqueValues(self, values):
        super().set_unique_values(values)
        if values is None:
            self._infoButton.hide()
        else:
            self._infoButton.show()

    def setDim(self, dim):
        assert isinstance(dim, Dimension)
        _kind = _METADATA_TYPES_I[dim.kind]
        idx = self._kindCB.findText(_kind)
        assert idx >= 0
        self._kindCB.setCurrentIndex(idx)
        idx = self._namesCB.findText(dim.name)
        if idx >= 0:
            self._namesCB.setCurrentIndex(idx)
        else:
            self._namesCB.addItem(dim.name)
        if dim.size is not None:
            self._sizeWidget.setValue(dim.size)
        if dim.range is not None:
            self._rangeWidget.setValue(dim.range)
        self._setTolerance(str(dim.tolerance))
        self.setUniqueValues(dim.unique_values)

    @property
    def _row(self):
        return self.__row

    @property
    def axis(self):
        return self._axis.value()

    @axis.setter
    def axis(self, axis):
        assert type(axis) is int
        self._axis.setValue(axis)

    @property
    def kind(self):
        assert self._kindCB.currentText() in _METADATA_TYPES
        return _METADATA_TYPES[self._kindCB.currentText()]

    @property
    def dim(self):
        return Dimension(
            name=self.name,
            kind=self.kind,
            size=self.size,
            _range=self.range,
            tolerance=self.tolerance,
        )

    @property
    def size(self):
        return self._sizeWidget.value

    @property
    def range(self):
        return self._rangeWidget.value

    @property
    def name(self):
        return self._namesCB.currentText()

    @property
    def tolerance(self):
        return float(self._toleranceLE.text())

    def _setSize(self, size):
        Dimension.set_size(self, size)
        self._sizeWidget.setValue(size=size)

    def _setRange(self, _range):
        Dimension.set_range(self, _range)
        self._rangeWidget.setValue(_range)

    def setNames(self, names):
        self._namesCB.clear()
        for name in sorted(list(names)):
            self._namesCB.addItem(name)

    def _setTolerance(self, tolerance):
        self._toleranceLE.setText(str(tolerance))

    def embedInTable(self, table, row):
        self.__row = row
        for column, widget in enumerate(
            (
                self._axis,
                self._kindCB,
                self._namesCB,
                self._sizeWidget,
                self._rangeWidget,
                self._toleranceLE,
                self._infoButton,
                self._rmButton,
            )
        ):
            table.setCellWidget(row, column, widget)

    def _updateNames(self, *args, **kwargs):
        """Update names for the current kind"""
        if self.__metadata is not None:
            _lastActiveName = self.name
            self.setNames(self.__metadata.get_keys(self.kind))
            idx = self._namesCB.findText(_lastActiveName)
            if idx >= 0:
                self._namesCB.setCurrentIndex(idx)

    def _setMetadata(self, metadata):
        self.__metadata = metadata
        self._updateNames()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "07/07/2021"


import enum
import numpy
import cv2
from scipy.ndimage import center_of_mass, fourier_shift
import skimage
from skimage import registration

from silx.utils.enum import Enum as _Enum

from darfix.io import utils
from .autofocus import normalized_variance


@enum.unique
class ShiftApproach(_Enum):
    """
    Different shifts approaches that can be used for the shift detection and correction.
    """

    LINEAR = "linear"
    FFT = "fft"


def compute_com(data):
    """
    Compute the center of mass of a stack of images.
    First it computes the intensity of every image (by summing its values), and then
    it uses scipy ``center_of_mass`` function to find the centroid.

    :param numpy.ndarray data: stack of images.
    :returns: the vector of intensities and the com.
    """
    intensity = data.sum(axis=1)
    return intensity, int(center_of_mass(intensity)[0])


def diff_com(img1, img2):
    """
    Finds the difference between the center of mass of two images.
    This can be used to find the shift between two images with distinguished
    center of mass.
    """
    return numpy.array(center_of_mass(img2)) - numpy.array(center_of_mass(img1))


def find_shift(img1, img2, upsample_factor=1000):
    """
    Uses the function ``register_translation`` from skimage to find the shift between two images.

    :param array_like img1: first image.
    :param array_like img2: second image, must be same dimensionsionality as ``img1``.
    :param int upsample_factor: optional.
    """
    if skimage.__version__ >= "0.19.0":
        return registration.phase_cross_correlation(
            img1,
            img2,
            upsample_factor=upsample_factor,
            return_error=False,
            normalization=None,
        )
    else:
        return registration.phase_cross_correlation(
            img1, img2, upsample_factor=upsample_factor, return_error=False
        )


def _numpy_fft_shift(img, dx, dy):
    """
    Shift an image by Fourier approach using Numpy library.

    :param array_like img: Image to shift
    :param number dx: shift in x axis
    :param number dy: shift in y axis
    :returns: ndarray
    """
    img = numpy.asarray(img, dtype=numpy.float32)
    xnum, ynum = img.shape
    """
    Create grid to apply the Fourier transformed theory for translated images.
    Each value of each matrix (Nx, Ny) correspons to the coordinate of the corresponding pixel.
    """
    Nx, Ny = numpy.meshgrid(numpy.fft.fftfreq(ynum), numpy.fft.fftfreq(xnum))
    c = numpy.fft.fft2(img) * numpy.exp(-2 * numpy.pi * 1j * (dx * Nx + dy * Ny))
    return numpy.fft.ifft2(c).real


def _opencv_fft_shift(img, dx, dy):
    """
    Shift an image by Fourier approach using opencv library.

    :param array_like img: Image to shift
    :param number dx: shift in x axis
    :param number dy: shift in y axis
    :returns: ndarray
    """
    img = numpy.asarray(img, dtype=numpy.float32)
    ynum, xnum = img.shape
    """
    Create grid to apply the Fourier transformed theory for translated images.
    Each value of each matrix (Nx, Ny) corresponds to the coordinate of the corresponding pixel
    divided by the size of the image.
    """
    Nx, Ny = numpy.meshgrid(numpy.fft.fftfreq(xnum), numpy.fft.fftfreq(ynum))
    dft = cv2.dft(img, flags=cv2.DFT_COMPLEX_OUTPUT)
    dft = dft[..., 0] + 1j * dft[..., 1]
    dft = dft * numpy.exp(-2 * numpy.pi * 1j * (dx * Nx + dy * Ny))
    real = numpy.reshape(dft.real, dft.real.shape + (1,))
    imag = numpy.reshape(dft.imag, dft.imag.shape + (1,))
    dft = numpy.concatenate((real, imag), axis=2)
    return cv2.idft(dft, flags=cv2.DFT_SCALE + cv2.DFT_REAL_OUTPUT)


def _scipy_fft_shift(img, dx, dy):
    """
    Shift an image by Fourier approach using scipy library.

    :param array_like img: Image to shift
    :param number dx: shift in x axis
    :param number dy: shift in y axis
    :returns: ndarray
    """
    return numpy.fft.ifftn(fourier_shift(numpy.fft.fftn(img), [dy, dx])).real


def apply_shift(img, shift, shift_approach="fft"):
    """
    Function to apply the shift to an image.

    :param 2-dimensional array_like img: Input array, can be complex.
    :param 2-dimensional array_like shift: The shift to be applied to the image. ``shift[0]``
        refers to the y-axis, ``shift[1]`` refers to the x-axis.
    :param Union[`linear`, `fft`] shift_approach: The shift method to be used to apply the shift.
    :returns: real ndarray
    """
    shift_approach = ShiftApproach.from_value(shift_approach)
    if shift_approach == ShiftApproach.LINEAR:
        img = img.astype(numpy.int16)
        M = numpy.float32([[1, 0, shift[1]], [0, 1, shift[0]]])
        return cv2.warpAffine(img, M, (img.shape[1], img.shape[0]))
    else:
        return _opencv_fft_shift(img, shift[1], shift[0])


def normalize(x):
    """
    Normalizes a vector or matrix.
    """
    if not numpy.count_nonzero(x):
        return x
    return x / numpy.linalg.norm(x)


def improve_linear_shift(
    data, v, h, epsilon, steps, nimages=None, shift_approach="linear"
):
    """
    Function to find the best shift between the images. It loops ``steps`` times,
    applying a different shift each, and trying to find the one that has the best result.

    :param array_like data: The stack of images.
    :param 2-dimensional array_like v: The vector with the direction of the shift.
    :param float epsilon: Maximum value of h
    :param int steps: Number of different tries of h.
    :param int nimages: The number of images to be used to find the best shift. It has to
        be smaller or equal as the length of the data. If it is smaller, the images used
        are chosen using `numpy.random.choice`, without replacement.
    :param Union[`linear`,`fft`] shift_approach: The shift method to be used to apply the shift.
    :returns: ndarray
    """
    shift_approach = ShiftApproach.from_value(shift_approach)
    v = numpy.asanyarray(v)
    iData = range(data.shape[0])

    if nimages:
        iData = numpy.random.choice(iData, nimages, False)

    score = {}
    utils.advancement_display(0, h + epsilon, "Finding shift")
    step = epsilon / steps
    for h_ in numpy.arange(0, h + epsilon, step):
        result = numpy.zeros(data[0].shape)
        for iFrame in iData:
            shift = h_ * v * iFrame
            result += apply_shift(data[iFrame], shift, shift_approach)

        # Compute score using normalized variance
        # TODO: add more autofocus options
        score[h_] = normalized_variance(result)
        utils.advancement_display(h_ + step, h + epsilon, "Finding shift")
    optimal_h = max(score.keys(), key=(lambda k: score[k]))
    return optimal_h


def shift_detection(data, steps, shift_approach="linear"):
    """
    Finds the linear shift from a set of images.

    :param ndarray data: Array with the images.
    :returns: A vector of length the number of images with the linear shift
        to apply to every image.
    :rtype: ndarray
    """
    if len(data) == 0:
        return []
    shape = data[0].shape
    first_sum = numpy.zeros(shape)
    second_sum = numpy.zeros(shape)
    for i in range(len(data)):
        if i < len(data) / 2:
            first_sum += data[i]
        else:
            second_sum += data[i]
    shift = find_shift(first_sum, second_sum, 1000)
    v = normalize(shift)
    v_ = 2 * shift / len(data)
    h = numpy.sqrt(v_[0] ** 2 + v_[1] ** 2)
    if not h:
        return numpy.outer([0, 0], numpy.arange(len(data)))
    epsilon = 2 * h
    h = improve_linear_shift(data, v, h, epsilon, steps, shift_approach=shift_approach)
    return numpy.outer(h * v, numpy.arange(len(data)))


def sum_images(image1, image2=None):
    if image2 is None:
        return numpy.asarray(image1)
    return numpy.asarray(image1) + numpy.asarray(image2)


def shift_correction(data, n_shift, shift_approach="fft", callback=None):
    """
    Function to apply shift correction technique to stack of images.

    :param array_like data: The stack of images to apply the shift to.
    :param array_like n_shift: Array with the shift to be applied at every image.
        The first row has the shifts in the y-axis and the second row the shifts
        in the x-axis. For image ``i`` the shift applied will be:
        ``shift_y = n_shift[0][i] shift_x = n_shift[1][i]```.
    :param Union[`linear`,`fft`] shift_approach: Name of the shift approach
        to be used. Default: `fft`.
    :param Union[None,Function] callback: Callback function to update the
        progress.
    :returns: The shifted images.
    :rtype: ndarray
    """
    shift_approach = ShiftApproach.from_value(shift_approach)

    shift = numpy.asanyarray(n_shift)

    if not numpy.any(shift):
        return data

    assert shift.shape == (
        2,
        len(data),
    ), "n_shift list has to be of same size as stack of images"

    shifted_data = numpy.empty(data.shape, dtype=numpy.float32)

    utils.advancement_display(0, len(data), "Applying shift")

    for count, frame in enumerate(data):
        shifted_data[count] = apply_shift(frame, shift[:, count], shift_approach)
        utils.advancement_display(count + 1, len(data), "Applying shift")
        if callback:
            callback(int(count / len(data) * 100))

    if data.dtype.kind == "f":
        if data.dtype.itemsize > shifted_data.dtype.itemsize:
            import warnings

            warnings.warn("Precision loss to float32")
    else:
        import warnings

        warnings.warn("Data cast to float32")

    return shifted_data


def random_search(data, optimal_shift, iterations, sigma=None, shift_approach="linear"):
    """
    Function that performs random search to a set of images to find an improved vector of shifts (one
    for each image). For this, it adds to the optimal shift a series of random samples obtained from
    a multivariate normal distribution, and selects the one with best score.

    :param array_like data: Set of images.
    :param ndarray optimal_shift: Array with two vectors with the linear shift found for axis y and x.
    :param int iterations: Number of times the random search should be done.
    :param number sigma: Standard deviation of the distribution.
    :param Union[`linear`,`fft`] shift_approach: Name of the shift approach
        to be used. Default: `linear`.
    :returns: A vector of length the number of images with the best shift found for every image.
    :rtype: ndarray
    """
    best_score = -numpy.inf
    best_result = numpy.empty(optimal_shift.shape)

    if sigma is None:
        # Using shift from second image
        sigma = abs(optimal_shift[:, 1] / 3)

    utils.advancement_display(0, iterations)
    for i in range(iterations):
        result = []
        normal = numpy.random.multivariate_normal(
            (0, 0), sigma * numpy.eye(2), len(data)
        ).T
        n_shift = optimal_shift + normal
        for iFrame in range(len(data)):
            shifty = n_shift[0][iFrame]
            shiftx = n_shift[1][iFrame]
            result += [apply_shift(data[iFrame], [shifty, shiftx], shift_approach)]
        score = normalized_variance(result)
        if best_score < score:
            best_result = n_shift
            best_score = score

        utils.advancement_display(i + 1, iterations)

    return best_result



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module contains library wide configuration.
"""

__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "19/11/2020"


class Config(object):
    """
    Class containing shared global configuration for the darfix project.

    .. versionadded:: 0.3
    """

    DEFAULT_COLORMAP_NAME = "cividis"
    DEFAULT_COLORMAP_NORM = "linear"
    """Default LUT for the plot widgets.

    The available list of names are available in the module
    :module:`silx.gui.colors`.

    .. versionadded:: 0.3
    """

    FWHM_VAL = 2.35482

    """Magic value that returns FWHM when multiplied by the standard deviation.

    .. versionadded:: 0.5
    """



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "15/09/2020"


import unittest
from silx.gui import qt
from orangecontrib.darfix.test.orangeWorkflowTest import OrangeWorkflowTest

try:
    from Orange.canvas.application.canvasmain import CanvasMainWindow
except ImportError:
    CanvasMainWindow = None
app = qt.QApplication.instance() or qt.QApplication([])


@unittest.skipUnless(
    CanvasMainWindow, "Only testing with orange versions older than 3.23.0"
)
class TestWorkflow(OrangeWorkflowTest):
    """
    Create workflow to test that the widgets are correctly displayed in the canvas.
    """

    def setUp(self):
        OrangeWorkflowTest.setUp(self)

        dataSelectionWidget = self.addWidget(
            "orangecontrib.darfix.widgets.dataselection.DataSelectionWidgetOW"
        )
        dataCopyWidget = self.addWidget(
            "orangecontrib.darfix.widgets.datacopy.DataCopyWidgetOW"
        )
        dataPartitionWidget = self.addWidget(
            "orangecontrib.darfix.widgets.datapartition.DataPartitionWidgetOW"
        )
        roiSelectionWidget = self.addWidget(
            "orangecontrib.darfix.widgets.roiselection.RoiSelectionWidgetOW"
        )
        noiseRemovalWidget = self.addWidget(
            "orangecontrib.darfix.widgets.noiseremoval.NoiseRemovalWidgetOW"
        )
        shiftCorrectionWidget = self.addWidget(
            "orangecontrib.darfix.widgets.shiftcorrection.ShiftCorrectionWidgetOW"
        )
        zsumWidget = self.addWidget("orangecontrib.darfix.widgets.zsum.ZSumWidgetOW")
        dimensionWidget = self.addWidget(
            "orangecontrib.darfix.widgets.dimensions.DimensionWidgetOW"
        )
        metadataWidget = self.addWidget(
            "orangecontrib.darfix.widgets.metadata.MetadataWidgetOW"
        )

        self.processOrangeEvents()

        self.link(dataSelectionWidget, "dataset", metadataWidget, "dataset")
        self.link(metadataWidget, "dataset", dataCopyWidget, "dataset")
        self.link(dataCopyWidget, "dataset", dataPartitionWidget, "dataset")
        self.link(dataPartitionWidget, "dataset", dimensionWidget, "dataset")
        self.link(dimensionWidget, "dataset", roiSelectionWidget, "dataset")
        self.link(roiSelectionWidget, "dataset", noiseRemovalWidget, "dataset")
        self.link(noiseRemovalWidget, "dataset", shiftCorrectionWidget, "dataset")
        self.link(shiftCorrectionWidget, "dataset", zsumWidget, "dataset")

        self.processOrangeEvents()

    def test(self):

        app.processEvents()

    def tearDown(self):
        OrangeWorkflowTest.tearDown(self)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "20/11/2020"

import logging
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Output

from silx.gui import qt

from darfix.gui.operationThread import OperationThread
from darfix.gui.datasetSelectionWidget import DatasetSelectionWidget
from darfix.core.process import DataSelection

_logger = logging.getLogger(__file__)


class DataSelectionWidgetOW(OWWidget):
    """
    Widget to select the data to be used in the dataset.
    """

    name = "data selection"
    icon = "icons/upload.svg"
    want_main_area = False
    ewokstaskclass = DataSelection

    # Outputs
    class Outputs:
        dataset = Output("dataset", tuple, bool)

    # Settings
    filenames = Setting(list(), schema_only=True)
    raw_filename = Setting(str(), schema_only=True)
    dark_filename = Setting(str(), schema_only=True)
    in_disk = Setting(bool(), schema_only=True)

    def __init__(self):
        super().__init__()

        self._widget = DatasetSelectionWidget()
        types = qt.QDialogButtonBox.Ok
        _buttons = qt.QDialogButtonBox(parent=self)
        _buttons.setStandardButtons(types)

        self.controlArea.layout().addWidget(self._widget)
        self.controlArea.layout().addWidget(_buttons)

        _buttons.accepted.connect(self._getDataset)

        self.__updatingData = False
        self.setDataset()

    def setDataset(self):
        self._widget.setRawFilenames(self.filenames)
        self._widget.setRawFilename(self.raw_filename)
        self._widget.setDarkFilename(self.dark_filename)
        if self.in_disk:
            self._widget._inDiskCB.setChecked(True)

    def _getDataset(self):
        if self.__updatingData:
            _logger.warning(
                "Another dataset is being loaded, please wait \
                            until it has finished"
            )
            return

        _logger.warning("create new dataset and emit them")
        # Create and start thread
        self._thread = OperationThread(self, self._widget.loadDataset)
        self._thread.finished.connect(self._sendSignal)
        self._thread.start()

        self.updateSettings()
        self.__updatingData = True
        self.information("Downloading dataset")

    def _updateDataset(self, widget, dataset):
        self._widget._updateDataset(dataset)
        self.Outputs.dataset.send(((self,) + self._widget.getDataset(), widget))

    def _sendSignal(self):
        """
        Function to emit the new dataset.
        Finishes the `downloading` state.
        """
        self._thread.finished.disconnect(self._sendSignal)
        self.__updatingData = False
        self.information()
        if self._thread.data:
            self.Outputs.dataset.send(((self,) + self._widget.getDataset(), None))
            self.close()

    def updateSettings(self):
        """
        Function to update the settings saved into the widget.
        """
        self.filenames = self._widget.getRawFilenames()
        self.raw_filename = self._widget.getRawFilename()
        self.dark_filename = self._widget.getDarkFilename()
        self.in_disk = self._widget._inDiskCB.isChecked()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "26/10/2020"


import logging
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Input, Output
from silx.gui import qt

from darfix.gui.rsmHistogramWidget import RSMHistogramWidget
from darfix.core.process import RSMHistogram

_logger = logging.getLogger(__file__)


class RSMHistogramWidgetOW(OWWidget):

    """
    Widget that sums a stack of images by the z axis.
    """

    name = "rsm histogram"
    icon = "icons/category.svg"
    want_main_area = False
    ewokstaskclass = RSMHistogram

    # Inputs
    class Inputs:
        dataset = Input("dataset", tuple)

    # Outputs
    class Outputs:
        dataset = Output("dataset", tuple)

    q = Setting(list())
    a = Setting(str())
    map_range = Setting(str())
    detector = Setting(str())
    units = Setting(str())
    n = Setting(list())
    map_shape = Setting(list())
    energy = Setting(float())

    def __init__(self):
        super().__init__()
        qt.QLocale.setDefault(qt.QLocale("en_US"))

        self._widget = RSMHistogramWidget(parent=self)
        self._widget.sigComputed.connect(self._compute)
        self.controlArea.layout().addWidget(self._widget)
        if self.q:
            self._widget.q = self.q
        if self.a:
            self._widget.a = self.a
        if self.map_range:
            self._widget.map_range = self.map_range
        if self.detector:
            self._widget.detector = self.detector
        if self.units:
            self._widget.units = self.units
        if self.n:
            self._widget.n = self.n
        if self.map_shape:
            self._widget.map_shape = self.map_shape
        if self.energy:
            self._widget.energy = self.energy

    @Inputs.dataset
    def setDataset(self, _input):
        if _input is not None:
            dataset, update = _input
            self._widget.setDataset(*dataset)
            if update is None:
                self.open()
            self.Outputs.dataset.send(((self,) + dataset[1:], update))

    def _updateDataset(self, widget, dataset):
        self._widget._updateDataset(widget, dataset)

    def _compute(self):
        self.q = self._widget.q.tolist()
        self.a = self._widget.a
        self.map_range = self._widget.map_range
        self.detector = self._widget.detector
        self.units = self._widget.units
        self.n = self._widget.n.tolist()
        self.map_shape = self._widget.map_shape.tolist()
        self.energy = self._widget.energy



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "25/04/2022"


import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.io.dictdump import dicttonx
from darfix.io.utils import create_nxdata_dict
from .rsmWidget import PixelSize
from .operationThread import OperationThread


class RSMHistogramWidget(qt.QMainWindow):
    """
    Widget to compute Reciprocal Space Map
    """

    sigComputed = qt.Signal()

    class _LineEditsWidget(qt.QWidget):
        def __init__(self, parent, dims=1, validator=None, placeholder=None):
            qt.QWidget.__init__(self, parent)
            self.setLayout(qt.QHBoxLayout())
            if placeholder:
                assert len(placeholder) == dims
            self._lineEdits = []
            for i in range(dims):
                lineEdit = qt.QLineEdit(parent=self)
                if placeholder:
                    lineEdit.setPlaceholderText(placeholder[i])
                if validator:
                    lineEdit.setValidator(validator)
                self._lineEdits.append(lineEdit)
                self.layout().addWidget(lineEdit)

            self.layout().setContentsMargins(0, 0, 0, 0)

        @property
        def value(self):
            values = []
            for le in self._lineEdits:
                values += [le.text() if le.text() != "" else None]
            return values

        def setValue(self, values):
            """
            :param int or None _range
            """
            assert type(values) is list

            for i, value in enumerate(values):
                self._lineEdits[i].setText(str(value))
                self._lineEdits[i].setCursorPosition(0)

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        self._idx = [(2, 1), (2, 0), (1, 0)]

        widget = qt.QWidget(parent=self)
        layout = qt.QGridLayout()
        label = qt.QLabel("Q:")
        font = label.font()
        font.setBold(True)
        label.setFont(font)
        layout.addWidget(label, 0, 0, 1, 1)
        self._q = self._LineEditsWidget(self, 3, qt.QIntValidator(), ["h", "k", "l"])
        self._q.setValue([1, 0, 1])
        layout.addWidget(self._q, 0, 1, 1, 2)
        label = qt.QLabel("Pseudocubic lattice:")
        label.setFont(font)
        layout.addWidget(label, 1, 0, 1, 2)
        self._a = qt.QLineEdit()
        self._a.setValidator(qt.QDoubleValidator())
        self._a.setText("4.08")
        layout.addWidget(self._a, 1, 1, 1, 2)
        label = qt.QLabel("Map range:")
        label.setFont(font)
        layout.addWidget(label, 2, 0, 1, 2)
        self._map_range = qt.QLineEdit()
        self._map_range.setValidator(qt.QDoubleValidator())
        self._map_range.setText("0.008")
        layout.addWidget(self._map_range, 2, 1, 1, 2)
        label = qt.QLabel("Detector:")
        label.setFont(font)
        layout.addWidget(label, 3, 0, 1, 2)
        self._detector = qt.QComboBox()
        self._detector.addItems(PixelSize.names())
        layout.addWidget(self._detector, 3, 1, 1, 2)
        label = qt.QLabel("Units:")
        label.setFont(font)
        layout.addWidget(label, 4, 0, 1, 2)
        self._units = qt.QComboBox()
        self._units.addItems(["Poulsen", "Gorfman"])
        layout.addWidget(self._units, 4, 1, 1, 2)
        label = qt.QLabel("n:")
        label.setFont(font)
        layout.addWidget(label, 5, 0, 1, 1)
        self._n = self._LineEditsWidget(self, 3, qt.QIntValidator(), ["h", "k", "l"])
        self._n.setValue([0, 1, 0])
        layout.addWidget(self._n, 5, 1, 1, 2)
        label = qt.QLabel("Map shape:")
        label.setFont(font)
        layout.addWidget(label, 6, 0, 1, 1)
        self._map_shape = self._LineEditsWidget(
            self, 3, qt.QIntValidator(), ["x", "y", "z"]
        )
        self._map_shape.setValue([200, 200, 200])
        layout.addWidget(self._map_shape, 6, 1, 1, 2)
        label = qt.QLabel("Energy:")
        label.setFont(font)
        layout.addWidget(label, 7, 0, 1, 2)
        self._energy = qt.QLineEdit()
        self._energy.setValidator(qt.QDoubleValidator())
        self._energy.setText("17")
        layout.addWidget(self._energy, 7, 1, 1, 2)
        self._computeB = qt.QPushButton("Compute")
        layout.addWidget(self._computeB, 8, 2, 1, 1)
        self._computeB.clicked.connect(self._computeRSM)
        self._plotWidget = qt.QWidget()
        self._plotWidget.setLayout(qt.QHBoxLayout())
        layout.addWidget(self._plotWidget, 9, 0, 1, 3)
        self._plotWidget.hide()
        self._exportButton = qt.QPushButton("Export maps")
        self._exportButton.hide()
        layout.addWidget(self._exportButton, 10, 2, 1, 1)
        self._exportButton.clicked.connect(self.exportMaps)

        widget.setLayout(layout)
        widget.setMinimumWidth(650)
        self.setCentralWidget(widget)

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter.

        :param Dataset dataset: dataset
        """
        self._parent = parent
        self.dataset = dataset
        self.indices = indices
        self._update_dataset = dataset
        self._thread = OperationThread(self, self.dataset.apply_fit)
        for i in reversed(range(self._plotWidget.layout().count())):
            self._plotWidget.layout().itemAt(i).widget().setParent(None)

        self._plotWidget.hide()
        self._exportButton.hide()

        self._plots = []
        for i in range(len(self._idx)):
            self._plots += [Plot2D(parent=self)]
            self._plots[-1].setDefaultColormap(Colormap(name="viridis"))
            self._plotWidget.layout().addWidget(self._plots[-1])

        self._thread = OperationThread(self, self.dataset.compute_rsm)

    @property
    def q(self):
        return numpy.array(self._q.value, dtype=int)

    @q.setter
    def q(self, q):
        self._q.setValue(q)

    @property
    def a(self):
        return float(self._a.text())

    @a.setter
    def a(self, a):
        self._a.setText(str(a))

    @property
    def map_range(self):
        return float(self._map_range.text())

    @map_range.setter
    def map_range(self, map_range):
        self._map_range.setText(str(map_range))

    @property
    def detector(self):
        return self._detector.currentText()

    @detector.setter
    def detector(self, detector):
        self._detector.setCurrentText(detector)

    @property
    def units(self):
        return self._units.currentText()

    @units.setter
    def units(self, units):
        self._units.setCurrentText(units)

    @property
    def n(self):
        return numpy.array(self._n.value, dtype=int)

    @n.setter
    def n(self, n):
        self._n.setValue(n)

    @property
    def map_shape(self):
        return numpy.array(self._map_shape.value, dtype=int)

    @map_shape.setter
    def map_shape(self, map_shape):
        self._map_shape.setValue(map_shape)

    @property
    def energy(self):
        return float(self._energy.text())

    @energy.setter
    def energy(self, energy):
        self._energy.setText(str(energy))

    def _computeRSM(self):

        self._computeB.setEnabled(False)
        self._thread.setArgs(
            Q=self.q,
            a=self.a,
            map_range=self.map_range,
            pixel_size=numpy.array(PixelSize[self.detector].value, dtype=float),
            units=self.units.lower(),
            n=self.n,
            map_shape=self.map_shape,
            energy=self.energy,
        )
        self.sigComputed.emit()
        self._thread.finished.connect(self._updateData)
        self._thread.start()

    def _updateData(self):
        self._thread.finished.disconnect(self._updateData)
        self._computeB.setEnabled(True)

        if self._thread.data is not None:
            arr, edges = self._thread.data

            arr = numpy.nan_to_num(arr)

            self._projections = []

            if self._units.currentText().lower() == "poulsen":
                self.labels = [r"$q_{rock}$", r"$q_{\perp}$", r"$q_{||}$"]
            else:
                self.labels = ["h", "k", "l"]

            for idx, (i, j) in enumerate(self._idx):
                self._projections += [numpy.sum(arr, axis=idx)]
                xscale = (edges[i][-1] - edges[i][0]) / len(edges[i])
                yscale = (edges[j][-1] - edges[j][0]) / len(edges[j])
                self._plots[idx].addImage(
                    self._projections[-1],
                    scale=(xscale, yscale),
                    origin=(edges[i][0], edges[j][0]),
                    xlabel=self.labels[i],
                    ylabel=self.labels[j],
                )

            self.edges = edges

            self._plotWidget.show()
            self._exportButton.show()

        else:
            print("\nComputation aborted")

    def exportMaps(self):
        """
        Creates dictionay with maps information and exports it to a nexus file
        """
        entry = "entry"

        nx = {
            entry: {"@NX_class": "NXentry"},
            "@NX_class": "NXroot",
            "@default": "entry",
        }

        for idx, (j, i) in enumerate(self._idx):
            axis = [self.edges[i][:-1], self.edges[j][:-1]]
            nx["entry"][str(idx)] = create_nxdata_dict(
                self._projections[idx],
                self.labels[i] + "_" + self.labels[j],
                axis,
                axes_names=[self.labels[i], self.labels[j]],
            )

        fileDialog = qt.QFileDialog()

        fileDialog.setFileMode(fileDialog.AnyFile)
        fileDialog.setAcceptMode(fileDialog.AcceptSave)
        fileDialog.setOption(fileDialog.DontUseNativeDialog)
        fileDialog.setDefaultSuffix(".h5")
        if fileDialog.exec_():
            dicttonx(nx, fileDialog.selectedFiles()[0])



# import sysconfig

# Category metadata.

# Category icon show in the menu
ICON = "icons/darfix_icon8.png"

# Background color for category background in menu
# and widget icon background in workflow.
BACKGROUND = "light-blue"

# Location of widget help files.
WIDGET_HELP_PATH = (
    # Development documentation
    # You need to build help pages manually using
    # make htmlhelp
    # inside doc folder
    # ("{DEVELOP_ROOT}/doc/build/htmlhelp/index.html", None),
    ("{DEVELOP_ROOT}/build/sphinx/htmlhelp/index.html", None),
    # Documentation included in wheel
    # Correct DATA_FILES entry is needed in setup.py and documentation has to be built
    # before the wheel is created.
    # ("{}/help/darfix/index.html".format(sysconfig.get_path("data")), None),
    # Online documentation url, used when the local documentation is not available.
    # Url should point to a page with a section Widgets. This section should
    # includes links to documentation pages of each widget. Matching is
    # performed by comparing link caption to widget name.
    # ("http://orange3-example-addon.readthedocs.io/en/latest/", "")
)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "28/02/2020"

import cv2
import numpy

from enum import Enum

from silx.image import sift


class Method(Enum):
    """
    Methods available to compute the matching.
    """

    orb_feature_matching = "orb feature matching"
    sift_feature_matching = "sift feature matching"
    euclidean_distance = "euclidean distance"

    @staticmethod
    def values():
        return list(map(lambda c: c.value, Method))


class ComponentsMatching:
    """
    Class to compute component matching.

    :param array_like components: List of stack of images. Every element of the
        list contains a stack of components from a certain dataset.
    """

    def __init__(self, components):

        self.components = components

    def _create_descriptors(self):
        """
        Function that detects and computes the keypoints and descriptors for
        the components.
        """
        orb = cv2.ORB_create()
        descripted_components = []

        for array in self.components:
            components = []
            for image in array:
                cv2.normalize(image, image, 0, 255, cv2.NORM_MINMAX)
                image = image.astype(numpy.uint8)
                kp, des = orb.detectAndCompute(image, None)
                components.append(Component(image, kp, des))
            descripted_components.append(components)

        return descripted_components

    def _create_sift_keypoints(self):

        keypoints = []

        for array in self.components:
            sift_ocl = sift.SiftPlan(template=array[0], devicetype="CPU")
            components = [sift_ocl(image) for image in array]
            keypoints.append(components)

        return keypoints

    def euclidean_distance(self, X, Y):
        """
        Compute euclidean distance between two images.
        """
        assert (
            X.shape == Y.shape
        ), "Images have to have same shape to compute euclidean distance"
        dst = numpy.linalg.norm(X - Y)  # their euclidean distances
        return dst

    def match_components(
        self, id1=None, id2=None, method=Method.orb_feature_matching, tol=8
    ):
        """
        Match components. Given the components x1,...,xn of dataset 1 and the
        components y1,...,ym of dataset 2, this function computes the pairs
        (xi,yi) that have better matching. Considering that each component of
        dataset 1 corresponds to one and only one component of dataset 2.

        :param Union[int,None] id1: Id of the first dataset to compare.
        :param Union[int,None] id2: Id of the second dataset to compare.
        :param Method method: Method to use for the matching.

        :returns: Dictionary with components ids of id1 per keys and their
            corresponding id component of id2 match per values, and dictionary
            with the matching info per pair of components.
        :rtype: (dict, dict)
        """
        if id1 is None or id2 is None:
            assert len(self.components) == 2, "Index of components must be given"
            id1 = 0
            id2 = 1

        good = {}
        final_matches = {}
        if method == Method.orb_feature_matching:
            self.descriptors = self._create_descriptors()

            bf = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)
            # Match components with id1 and id2
            for i, component1 in enumerate(self.descriptors[id1]):
                if component1.descriptor is not None:
                    for j, component2 in enumerate(self.descriptors[id2]):
                        if component2.descriptor is not None:

                            # Match descriptors
                            good[(i, j)] = numpy.array(
                                bf.match(component1.descriptor, component2.descriptor)
                            )

            best_v = []
            # Add matches sorted by number of matches found.
            for x, y in sorted(good, key=lambda match: len(good[match]), reverse=True):
                # Only add match if neither x nor y are already in the list.
                if x not in final_matches.keys() and y not in final_matches.values():

                    kp1 = []
                    kp2 = []
                    for match in good[(x, y)]:
                        kp1 += [self.descriptors[id1][x].keypoints[match.queryIdx].pt]
                        kp2 += [self.descriptors[id2][y].keypoints[match.trainIdx].pt]
                    if len(kp1) > 1 and len(kp2) > 1:
                        v = numpy.mean(numpy.array(kp2) - numpy.array(kp1), axis=0)
                    else:
                        v = numpy.array(kp2) - numpy.array(kp1)
                    if not numpy.any(best_v):
                        best_v = v
                        final_matches[x] = y
                    elif numpy.linalg.norm(best_v - v) < tol:
                        final_matches[x] = y

        elif method == Method.sift_feature_matching:
            keypoints = self._create_sift_keypoints()
            best_v = []
            mp = sift.MatchPlan()
            # Match components with id1 and id2
            for i, kp1 in enumerate(keypoints[id1]):
                for j, kp2 in enumerate(keypoints[id2]):
                    # Match descriptors
                    good[(i, j)] = mp.match(kp1, kp2)
            # Add matches sorted by number of matches found.
            for x, y in sorted(
                good, key=lambda match: good[match].shape[0], reverse=True
            ):
                # Only add match if neither x nor y are already in the list.
                if x not in final_matches.keys() and y not in final_matches.values():
                    v = numpy.array(
                        [
                            numpy.median(good[(x, y)][:, 1].x - good[(x, y)][:, 0].x),
                            numpy.median(good[(x, y)][:, 1].y - good[(x, y)][:, 0].y),
                        ]
                    )
                    if not numpy.any(best_v):
                        best_v = v
                        final_matches[x] = y
                    elif numpy.linalg.norm(best_v - v) < tol:
                        final_matches[x] = y

        elif method == Method.euclidean_distance:
            for i, X in enumerate(self.components[id1]):
                for j, Y in enumerate(self.components[id2]):
                    good[(i, j)] = self.euclidean_distance(X, Y)
            # Add matches sorted by distance.
            for x, y in sorted(good, key=lambda match: good[match]):
                # Only add match if neither x nor y are already in the list.
                if x not in final_matches.keys() and y not in final_matches.values():
                    final_matches[x] = y
        return final_matches, good

    def draw_matches(
        self, final_matches, matches, id1=None, id2=None, displayMatches=False
    ):
        """
        Create stack of images with each pair of matches.

        :param dict final_matches: Dictionary with the best pairs of matches per items.
        :param dict matches: Dictionary with keys the pairs of matches and with
            values the information of every pair of components.
        :param Union[int,None] id1: Id of the first dataset to compare.
        :param Union[int,None] id2: Id of the second dataset to compare.
        :param bool displayMatches: If True, dictionary `matches` has to contain
            values of type `cv2.DMatch`.

        :returns array_like: stack with the pairs of images, and if so, info
            about the matching.
        """
        if id1 is None or id2 is None:
            assert len(self.components) == 2, "Index of components must be given"
            id1 = 0
            id2 = 1
        stack = []
        for i, img1 in enumerate(self.components[id1]):
            if i in final_matches:
                j = final_matches[i]
                img2 = self.components[id2][j]
                # Show link between features
                if displayMatches:
                    # Check that all values are of type `cv2.DMatch`
                    assert all(
                        (isinstance(match, cv2.DMatch) for match in values)
                        for values in matches.values()
                    ), "Dictionary `matches` has to contain values of type `cv2.DMatch`"

                    img = cv2.drawMatches(
                        self.descriptors[id1][i].image,
                        self.descriptors[id1][i].keypoints,
                        self.descriptors[id2][j].image,
                        self.descriptors[id2][j].keypoints,
                        matches[(i, j)],
                        None,
                        flags=2,
                    )
                    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
                else:
                    shape1, shape2 = img1.shape, img2.shape
                    img = numpy.zeros(
                        (max(shape1[0], shape2[0]), shape1[1] + shape2[1])
                    )
                    img[: shape1[0], : shape1[1]] = img1
                    img[: shape2[0], shape1[1] :] = img2
            else:
                shape1 = img1.shape
                shape2 = self.components[id2][0].shape
                img = numpy.zeros((max(shape1[0], shape2[0]), shape1[1] + shape2[1]))
                img[: shape1[0], : shape1[1]] = img1
            stack.append(img)
        return stack


class Component:
    """
    Class Component. Describes a component of a dataset (image) with its keypoints
    and descriptors.
    """

    def __init__(self, image, kp, des):
        self._image = image
        self._keypoints = kp
        self._descriptor = des

    @property
    def keypoints(self):
        return self._keypoints

    @property
    def descriptor(self):
        return self._descriptor

    @property
    def image(self):
        return self._image



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

"""
Module for defining processes to be used by the library `ewoks`. Each of
the processes defined here can be used (its corresponding widgets) within an
Orange workflow and later be converted to an Ewoks workflow without the GUI part needed.
"""

__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "16/07/2021"


import os
from collections import Mapping
from typing import Iterable, Union, Optional
from ewoksutils.import_utils import qualname
import numpy
import copy

from silx.gui import qt
from darfix.core import utils
from darfix.gui.blindSourceSeparationWidget import Method
from darfix.gui.grainPlotWidget import GrainPlotWidget
from darfix.gui.rsmHistogramWidget import RSMHistogramWidget
from darfix.gui.zSumWidget import ZSumWidget
from darfix.gui.rsmWidget import PixelSize
from darfix.core.data_selection import load_process_data
from darfix.io.utils import write_components
from ewokscore import Task
from ewokscore.graph import TaskGraph


def graph_data_selection(
    graph: TaskGraph,
    filenames: Union[str, Iterable[str]],
    root_dir: Optional[str] = None,
    in_memory: bool = True,
    dark_filename: Optional[str] = None,
    copy_files: bool = True,
):
    task_identifier = qualname(DataSelection)
    for node_attrs in graph.graph.nodes.values():
        if node_attrs.get("task_identifier") == task_identifier:
            default_inputs = node_attrs["default_inputs"] = list()
            default_inputs.append({"name": "root_dir", "value": root_dir})
            default_inputs.append({"name": "filenames", "value": filenames})
            default_inputs.append({"name": "dark_filename", "value": dark_filename})
            default_inputs.append({"name": "copy_files", "value": copy_files})
            default_inputs.append({"name": "in_memory", "value": in_memory})
            return
    raise RuntimeError(f"Workflow {graph} does not contain a 'DataSelection' task")


class DataSelection(
    Task,
    input_names=["filenames"],
    optional_input_names=["root_dir", "in_memory", "dark_filename", "copy_files"],
    output_names=["dataset"],
):
    """Simple util class to ignore a processing when executing without GUI"""

    def run(self):
        in_memory = self.inputs.in_memory
        if in_memory == self.MISSING_DATA:
            in_memory = True
        copy_files = self.inputs.copy_files
        if copy_files == self.MISSING_DATA:
            copy_files = True
        dark_filename = self.inputs.dark_filename
        if dark_filename == self.MISSING_DATA:
            dark_filename = None
        self.outputs.dataset = load_process_data(
            filenames=self.inputs.filenames,
            root_dir=self.inputs.root_dir,
            dark_filename=dark_filename,
            in_memory=in_memory,
            copy_files=copy_files,
        )


class DataCopy(
    Task,
    input_names=["dataset"],
    output_names=["dataset"],
):
    def run(self):
        self.outputs.dataset = copy.deepcopy(self.inputs.dataset)


class DataPassThrough(
    Task,
    input_names=["dataset"],
    output_names=["dataset"],
):
    def run(self):
        self.outputs.dataset = self.inputs.dataset


class NoiseRemoval(
    Task,
    input_names=["dataset"],
    optional_input_names=[
        "method",
        "background_type",
        "step",
        "chunks",
        "kernel_size",
        "bottom_threshold",
        "mask",
    ],
    output_names=["dataset"],
):
    def run(self):
        dataset, indices, li_indices, bg_dataset = self.inputs.dataset
        if self.inputs.method:
            step = int(self.inputs.step) if self.inputs.step else None
            chunks = self.inputs.chunks if self.inputs.chunks else None

            bg = None
            if self.inputs.background_type == "Dark data":
                bg = bg_dataset
            elif self.inputs.background_type == "Low intensity data":
                bg = li_indices

            dataset = dataset.apply_background_subtraction(
                indices=indices,
                method=self.inputs.method,
                background=bg,
                step=step,
                chunk_shape=chunks,
            )
        if self.inputs.kernel_size:
            dataset = dataset.apply_hot_pixel_removal(
                indices=indices, kernel=int(self.inputs.kernel_size)
            )

        if self.inputs.bottom_threshold:
            dataset = dataset.apply_threshold_removal(
                bottom=int(self.inputs.bottom_threshold)
            )

        if self.inputs.mask:
            dataset = dataset.apply_mask_removal(mask=self.inputs.mask)

        self.outputs.dataset = dataset, indices, li_indices, bg_dataset


class RoiSelection(
    Task,
    input_names=["dataset"],
    optional_input_names=["roi_origin", "roi_size"],
    output_names=["dataset"],
):
    def run(self):
        dataset, indices, li_indices, bg_dataset = self.inputs.dataset
        origin = numpy.flip(self.inputs.roi_origin) if self.inputs.roi_origin else []
        size = numpy.flip(self.inputs.roi_size) if self.inputs.roi_size else []
        if len(origin) and len(size):
            dataset = dataset.apply_roi(origin=origin, size=size)
            if bg_dataset:
                bg_dataset = bg_dataset.apply_roi(origin=origin, size=size)
        self.outputs.dataset = dataset, indices, li_indices, bg_dataset


class DataPartition(
    Task,
    input_names=["dataset"],
    optional_input_names=["bins", "n_bins"],
    output_names=["dataset"],
):
    def run(self):
        dataset, indices, li_indices, bg_dataset = self.inputs.dataset
        bins = self.inputs.bins if self.inputs.bins else None
        nbins = self.inputs.n_bins if self.inputs.n_bins else 1
        indices, li_indices = dataset.partition_by_intensity(bins, nbins)
        self.outputs.dataset = dataset, indices, li_indices, bg_dataset


class DimensionDefinition(
    Task, input_names=["dataset", "_dims"], output_names=["dataset"]
):
    def run(self):
        dataset, indices, li_indices, bg_dataset = self.inputs.dataset
        assert isinstance(self.inputs._dims, Mapping)
        dims = utils.convertDictToDim(self.inputs._dims)
        if dataset is not None and len(dataset.data.metadata) > 0:
            for axis, dim in dims.items():
                assert type(axis) is int
                dataset.add_dim(axis=axis, dim=dim)
            try:
                dataset = dataset.reshape_data()
                for axis, dimension in dataset.dims:
                    if dataset.dims.ndim > 1:
                        metadata = numpy.swapaxes(dataset.data.metadata, 0, axis)[0]
                    else:
                        metadata = dataset.data.metadata
                    values = []
                    for data in metadata:
                        try:
                            values += [
                                data.get_value(
                                    kind=dimension.kind, name=dimension.name
                                )[0]
                            ]
                        except KeyError:
                            values += ["0"]
                    dimension.set_unique_values(values)
            except ValueError:
                for axis, dimension in dataset.dims:
                    values = numpy.unique(
                        self.dataset.get_dimensions_values()[dimension.name]
                    )
                    dimension.set_unique_values(values)
                dataset = dataset.reshape_data()
        self.outputs.dataset = dataset, indices, li_indices, bg_dataset


class ShiftCorrection(
    Task,
    input_names=["dataset"],
    optional_input_names=["shift"],
    output_names=["dataset"],
):
    def run(self):
        dataset, indices, li_indices, bg_dataset = self.inputs.dataset
        if not self.inputs.shift:
            raise ValueError("Shift not defined")
        frames = numpy.arange(dataset.get_data(indices=indices).shape[0])
        dataset = dataset.apply_shift(
            numpy.outer(self.inputs.shift, frames), indices=indices
        )
        self.outputs.dataset = dataset, indices, li_indices, bg_dataset


class BlindSourceSeparation(
    Task,
    input_names=["dataset", "method"],
    optional_input_names=["n_comp"],
    output_names=["dataset"],
):
    def run(self):
        dataset, indices, li_indices, bg_dataset = self.inputs.dataset
        n_comp = self.inputs.n_comp if self.inputs.n_comp else None
        method = Method[self.inputs.method]
        if method == Method.PCA:
            comp, W = dataset.pca(n_comp, indices=indices)
        elif method == Method.NICA:
            comp, W = dataset.nica(n_comp, indices=indices)
        elif method == Method.NMF:
            comp, W = dataset.nmf(n_comp, indices=indices)
        elif method == Method.NICA_NMF:
            comp, W = dataset.nica_nmf(n_comp, indices=indices)
        else:
            raise ValueError("BSS method not managed")
        n_comp = comp.shape[0]
        shape = dataset.get_data()[0].shape
        comp = comp.reshape(n_comp, shape[0], shape[1])
        if li_indices is not None:
            # If filter data is activated, the matrix W has reduced dimensionality, so reshaping is not possible
            # Create empty array with shape the total number of frames
            W = numpy.zeros((dataset.nframes, n_comp))
            # Set actual values of W where threshold of filter is True
            W[indices] = W
            W = W
        write_components(
            os.path.join(dataset.dir, "components.h5"),
            "entry",
            dataset.get_dimensions_values(),
            W,
            comp,
            1,
        )
        self.outputs.dataset = dataset, indices, li_indices, bg_dataset


class RockingCurves(
    Task,
    input_names=["dataset"],
    optional_input_names=["int_thresh", "dimension"],
    output_names=["dataset"],
):
    def run(self):
        dataset, indices, li_indices, bg_dataset = self.inputs.dataset
        int_thresh = float(self.inputs.int_thresh) if self.inputs.int_thresh else None
        dataset = dataset.apply_fit(indices=indices, int_thresh=int_thresh)
        self.outputs.dataset = dataset, indices, li_indices, bg_dataset


class GrainPlot(Task, input_names=["dataset"], output_names=["dataset"]):
    def run(self):
        app = qt.QApplication([])
        widget = GrainPlotWidget()
        if self.inputs.dataset:
            widget.setDataset(*(None,) + self.inputs.dataset)
        widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        widget.show()
        app.exec_()
        self.outputs.dataset = self.inputs.dataset[1:]


class Transformation(
    Task,
    input_names=["dataset"],
    optional_input_names=[
        "magnification",
        "pixelSize",
        "kind",
        "rotate",
        "orientation",
    ],
    output_names=["dataset"],
):
    def run(self):
        dataset, indices, li_indices, bg_dataset = self.inputs.dataset
        magnification = self.inputs.magnification if self.inputs.magnification else None
        orientation = self.inputs.orientation if self.inputs.orientation else None
        pixelSize = self.inputs.pixelSize if self.inputs.pixelSize else None
        kind = self.inputs.kind if self.inputs.kind else None
        rotate = self.inputs.rotate if self.inputs.rotate else None
        if dataset and dataset.dims.ndim:
            if dataset.dims.ndim == 1 and kind:
                dataset.compute_transformation(
                    PixelSize[pixelSize].value, kind="rsm", rotate=rotate
                )
            else:
                if orientation == -1 or orientation is None:
                    dataset.compute_transformation(magnification, topography=[False, 0])
                else:
                    dataset.compute_transformation(
                        magnification, topography=[True, orientation]
                    )
        self.outputs.dataset = dataset, indices, li_indices, bg_dataset


class ZSum(
    Task,
    input_names=["dataset"],
    optional_input_names=["plot"],
    output_names=["dataset"],
):
    def run(self):
        self.outputs.dataset = self.inputs.dataset
        if self.inputs.plot:
            app = qt.QApplication([])
            widget = ZSumWidget()
            if self.inputs.dataset:
                widget.setDataset((None,) + self.inputs.dataset)
            widget.setAttribute(qt.Qt.WA_DeleteOnClose)
            widget.show()
            app.exec_()


class Projection(
    Task,
    input_names=["dataset"],
    optional_input_names=["dimension"],
    output_names=["dataset"],
):
    def run(self):
        dataset, indices, li_indices, bg_dataset = self.inputs.dataset
        dimension = self.inputs.dimension
        if dimension:
            dataset = dataset.project_data(dimension=dimension, indices=indices)
        self.outputs.dataset = dataset, indices, li_indices, bg_dataset


class RSMHistogram(
    Task,
    input_names=["dataset"],
    optional_input_names=[
        "q",
        "a",
        "map_range",
        "detector",
        "units",
        "n",
        "map_shape",
        "energy",
    ],
    output_names=["dataset"],
):
    def run(self):
        app = qt.QApplication([])
        widget = RSMHistogramWidget()
        if self.inputs.dataset:
            widget.setDataset(*(None,) + self.inputs.dataset)
        widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        # TODO: Only show computed maps?
        if self.inputs.q:
            widget.q = self.inputs.q
        if self.inputs.a:
            widget.a = self.inputs.a
        if self.inputs.map_range:
            widget.map_range = self.inputs.map_range
        if self.inputs.detector:
            widget.detector = self.inputs.detector
        if self.inputs.units:
            widget.units = self.inputs.units
        if self.inputs.n:
            widget.n = self.inputs.n
        if self.inputs.map_shape:
            widget.map_shape = self.inputs.map_shape
        if self.inputs.energy:
            widget.energy = self.inputs.energy
        widget.show()
        app.exec_()
        self.outputs.dataset = self.inputs.dataset[1:]



import os
from typing import Iterable, Optional, Union
from darfix.core.dataset import Dataset


def load_process_data(
    filenames: Union[str, Iterable[str]],
    root_dir: Optional[str] = None,
    in_memory: bool = True,
    dark_filename: Optional[str] = None,
    copy_files: bool = True,
    isH5: bool = False,
    title: str = "",
):
    """When `filenames` is a string, it will be treated as a file pattern."""
    indices = li_indices = None
    root_dir_specified = bool(root_dir)

    if isinstance(filenames, str):
        if not root_dir_specified:
            root_dir = os.path.dirname(filenames)
        dataset = Dataset(
            _dir=root_dir,
            first_filename=filenames,
            in_memory=in_memory,
            copy_files=copy_files,
            isH5=isH5,
            title=title,
        )
    else:
        filenames = list(filenames)
        if not root_dir_specified:
            root_dir = os.path.dirname(filenames[0])
        dataset = Dataset(
            _dir=root_dir,
            filenames=filenames,
            in_memory=in_memory,
            copy_files=copy_files,
            isH5=isH5,
            title=title,
        )

    if dark_filename:
        dark_root_dir = os.path.join(dataset.dir, "dark")
        if not os.path.isdir(dark_root_dir):
            os.mkdir(dark_root_dir)
        bg_dataset = Dataset(
            _dir=dark_root_dir,
            first_filename=dark_filename,
            copy_files=False,
            isH5=isH5,
        )
    else:
        bg_dataset = None

    return dataset, indices, li_indices, bg_dataset
